#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2021, Andrew Klychkov <andrew.a.klychkov@gmail.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = r'''
---
module: mysql_role

short_description: Adds, removes, or updates a MySQL or MariaDB role

description:
   - Adds, removes, or updates a MySQL or MariaDB role.
   - Roles are supported since MySQL 8.0.0 and MariaDB 10.0.5.

version_added: '2.2.0'

options:
  name:
    description:
      - Name of the role to add or remove.
    type: str
    required: true

  admin:
    description:
      - Supported by B(MariaDB).
      - Name of the admin user of the role (the I(login_user), by default).
    type: str

  priv:
    description:
      - "MySQL privileges string in the format: C(db.table:priv1,priv2)."
      - "You can specify multiple privileges by separating each one using
        a forward slash: C(db.table:priv/db.table:priv)."
      - The format is based on MySQL C(GRANT) statement.
      - Database and table names can be quoted, MySQL-style.
      - If column privileges are used, the C(priv1,priv2) part must be
        exactly as returned by a C(SHOW GRANT) statement. If not followed,
        the module will always report changes. It includes grouping columns
        by permission (C(SELECT(col1,col2)) instead of C(SELECT(col1),SELECT(col2))).
      - Can be passed as a dictionary (see the examples).
      - Supports GRANTs for procedures and functions
        (see the examples for the M(ansible.mysql.mysql_user) module).
    type: raw

  append_privs:
    description:
      - Append the privileges defined by the I(priv) option to the existing ones
        for this role instead of overwriting them. Mutually exclusive with I(subtract_privs).
    type: bool
    default: false

  subtract_privs:
    description:
      - Revoke the privileges defined by the I(priv) option and keep other existing privileges.
        If set, invalid privileges in I(priv) are ignored.
        Mutually exclusive with I(append_privs).
    version_added: '3.2.0'
    type: bool
    default: false

  members:
    description:
      - List of members of the role.
      - For users, use the format C(username@hostname).
        Always specify the hostname part explicitly.
      - For roles, use the format C(rolename).
      - Mutually exclusive with I(admin).
    type: list
    elements: str

  append_members:
    description:
      - Add members defined by the I(members) option to the existing ones
        for this role instead of overwriting them.
      - Mutually exclusive with the I(detach_members) and I(admin) option.
    type: bool
    default: false

  detach_members:
    description:
      - Detaches members defined by the I(members) option from the role
        instead of overwriting all the current members.
      - Mutually exclusive with the I(append_members) and I(admin) option.
    type: bool
    default: false

  set_default_role_all:
    description:
      - Is not supported by MariaDB and is silently ignored when working with MariaDB.
      - If C(yes), runs B(SET DEFAULT ROLE ALL TO) each of the I(members) when changed.
      - If you want to avoid this behavior, set this option to C(no) explicitly.
    type: bool
    default: true

  state:
    description:
      - If C(present) and the role does not exist, creates the role.
      - If C(present) and the role exists, does nothing or updates its attributes.
      - If C(absent), removes the role.
    type: str
    choices: [ absent, present ]
    default: present

  check_implicit_admin:
    description:
      - Check if mysql allows login as root/nopassword before trying supplied credentials.
      - If success, passed I(login_user)/I(login_password) will be ignored.
    type: bool
    default: false

  members_must_exist:
    description:
      - When C(yes), the module fails if any user in I(members) does not exist.
      - When C(no), users in I(members) which don't exist are simply skipped.
    type: bool
    default: true

  column_case_sensitive:
    description:
      - The default is C(true).
      - When C(true), the module will not uppercase the field in the privileges.
      - When C(false), the field names will be upper-cased. This was the default before this
        feature was introduced but since MySQL/MariaDB is case sensitive you should set this
        to C(true) in most cases.
      - The default changed from C(false) to C(true) in v4.0.0
    type: bool
    default: true
    version_added: '3.8.0'

  sql_log_bin:
    description:
      - Whether binary logging should be enabled or disabled for the connection.
    type: bool
    default: true
    version_added: '4.1.0'

notes:
  - Roles are supported since MySQL 8.0.0 and MariaDB 10.0.5.
  - Pay attention that the module runs C(SET DEFAULT ROLE ALL TO)
    all the I(members) passed by default when the state has changed.
    If you want to avoid this behavior, set I(set_default_role_all) to C(no).

attributes:
  check_mode:
    support: full
  idempotent:
    support: full

seealso:
  - module: ansible.mysql.mysql_user
  - name: MySQL role reference
    description: Complete reference of the MySQL role documentation.
    link: https://dev.mysql.com/doc/refman/8.0/en/create-role.html

author:
  - Andrew Klychkov (@Andersson007)
  - Felix Hamme (@betanummeric)
  - kmarse (@kmarse)
  - Laurent Indermühle (@laurent-indermuehle)

extends_documentation_fragment:
  - ansible.mysql.mysql
'''

EXAMPLES = r'''
# If you encounter the "Please explicitly state intended protocol" error,
# use the login_unix_socket argument, for example, login_unix_socket: /run/mysqld/mysqld.sock

# Example of a .my.cnf file content for setting a root password
# [client]
# user=root
# password=n<_665{vS43y
#
# Example of a privileges dictionary passed through the priv option
# priv:
#   'mydb.*': 'INSERT,UPDATE'
#   'anotherdb.*': 'SELECT'
#   'yetanotherdb.*': 'ALL'
#
# You can also use the string format like in the ansible.mysql.mysql_user module, for example
# mydb.*:INSERT,UPDATE/anotherdb.*:SELECT/yetanotherdb.*:ALL
#
# For more examples on how to specify privileges, refer to the ansible.mysql.mysql_user module

# Create a role developers with all database privileges
# and add alice and bob as members.
# The statement 'SET DEFAULT ROLE ALL' to them will be run.
- name: Create role developers, add members
  ansible.mysql.mysql_role:
    name: developers
    state: present
    priv: '*.*:ALL'
    members:
      - 'alice@%'
      - 'bob@%'

- name: Same as above but do not run SET DEFAULT ROLE ALL TO each member
  ansible.mysql.mysql_role:
    name: developers
    state: present
    priv: '*.*:ALL'
    members:
      - 'alice@%'
      - 'bob@%'
    set_default_role_all: false

# Assuming that the role developers exists,
# add john to the current members
- name: Add members to an existing role
  ansible.mysql.mysql_role:
    name: developers
    state: present
    append_members: true
    members:
      - 'joe@localhost'

# Create role readers with the SELECT privilege
# on all tables in the fiction database
- name: Create role developers, add members
  ansible.mysql.mysql_role:
    name: readers
    state: present
    priv: 'fiction.*:SELECT'

# Assuming that the role readers exists,
# add the UPDATE privilege to the role on all tables in the fiction database
- name: Create role developers, add members
  ansible.mysql.mysql_role:
    name: readers
    state: present
    priv: 'fiction.*:UPDATE'
    append_privs: true

- name: Create role with the 'SELECT' and 'UPDATE' privileges in db1 and db2
  ansible.mysql.mysql_role:
    state: present
    name: foo
    priv:
      'db1.*': 'SELECT,UPDATE'
      'db2.*': 'SELECT,UPDATE'

- name: Remove joe from readers
  ansible.mysql.mysql_role:
    state: present
    name: readers
    members:
      - 'joe@localhost'
    detach_members: true

- name: Remove the role readers if exists
  ansible.mysql.mysql_role:
    state: absent
    name: readers

- name: Example of using login_unix_socket to connect to the server
  ansible.mysql.mysql_role:
    name: readers
    state: present
    login_unix_socket: /var/run/mysqld/mysqld.sock

# Pay attention that the admin cannot be changed later
# and will be ignored if a role currently exists.
# To change members, you need to run a separate task using the admin
# of the role as the login_user.
- name: On MariaDB, create the role readers with alice as its admin
  ansible.mysql.mysql_role:
    state: present
    name: readers
    admin: 'alice@%'

- name: Create the role business, add the role marketing to members
  ansible.mysql.mysql_role:
    state: present
    name: business
    members:
      - marketing

- name: Ensure the role foo does not have the DELETE privilege
  ansible.mysql.mysql_role:
    state: present
    name: foo
    subtract_privs: true
    priv:
      'db1.*': DELETE

- name: Add some members to a role and skip not-existent users
  ansible.mysql.mysql_role:
    state: present
    name: foo
    append_members: true
    members_must_exist: false
    members:
      - 'existing_user@localhost'
      - 'not_existing_user@localhost'

- name: Detach some members from a role and ignore not-existent users
  ansible.mysql.mysql_role:
    state: present
    name: foo
    detach_members: true
    members_must_exist: false
    members:
      - 'existing_user@localhost'
      - 'not_existing_user@localhost'

- name: Create role without binary logging
  ansible.mysql.mysql_role:
    name: readers
    state: present
    priv: 'fiction.*:SELECT'
    sql_log_bin: false
'''

RETURN = '''#'''

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.ansible.mysql.plugins.module_utils.mysql import (
    mysql_connect,
    mysql_driver,
    mysql_driver_fail_msg,
    mysql_common_argument_spec
)
from ansible_collections.ansible.mysql.plugins.module_utils.user import (
    convert_priv_dict_to_str,
    get_user_implementation,
    get_mode,
    user_mod,
    privileges_grant,
    privileges_unpack,
)
from ansible.module_utils.common.text.converters import to_native


def normalize_users(module, users, is_mariadb=False):
    """Normalize passed user names.

    Example of transformation:
    ['user0'] => [('user0', '')] / ['user0'] => [('user0', '%')]
    ['user0@host0'] => [('user0', 'host0')]

    Args:
        module (AnsibleModule): Object of the AnsibleModule class.
        users (list): List of user names.
        is_mariadb (bool): Flag indicating we are working with MariaDB

    Returns:
        list: List of tuples like [('user0', ''), ('user0', 'host0')].
    """
    normalized_users = []

    for user in users:
        try:
            tmp = user.split('@')

            if tmp[0] == '':
                module.fail_json(msg="Member's name cannot be empty.")

            if len(tmp) == 1:
                if not is_mariadb:
                    normalized_users.append((tmp[0], '%'))
                else:
                    normalized_users.append((tmp[0], ''))

            elif len(tmp) == 2:
                normalized_users.append((tmp[0], tmp[1]))

        except Exception as e:
            msg = ('Error occured while parsing the name "%s": %s. '
                   'It must be in the format "username" or '
                   '"username@hostname" ' % (user, to_native(e)))
            module.fail_json(msg=msg)

    return normalized_users


class DbServer():
    """Class to fetch information from a database.

    Args:
        module (AnsibleModule): Object of the AnsibleModule class.
        cursor (cursor): Cursor object of a database Python connector.

    Attributes:
        module (AnsibleModule): Object of the AnsibleModule class.
        cursor (cursor): Cursor object of a database Python connector.
        role_impl (library): Corresponding library depending
            on a server type (MariaDB or MySQL)
        mariadb (bool): True if MariaDB, False otherwise.
        roles_supported (bool): True if roles are supported, False otherwise.
        users (set): Set of users existing in a DB in the form (username, hostname).
    """
    def __init__(self, module, cursor):
        self.module = module
        self.cursor = cursor
        self.role_impl = self.get_implementation()
        self.mariadb = self.role_impl.is_mariadb()
        self.roles_supported = self.role_impl.supports_roles(self.cursor)
        self.users = set(self.__get_users())

    def is_mariadb(self):
        """Get info whether a DB server is a MariaDB instance.

        Returns:
            self.mariadb: Attribute value.
        """
        return self.mariadb

    def supports_roles(self):
        """Get info whether a DB server supports roles.

        Returns:
            self.roles_supported: Attribute value.
        """
        return self.roles_supported

    def get_implementation(self):
        """Get a current server implementation depending on its type.

        Returns:
            library: Depending on a server type (MySQL or MariaDB).
        """
        self.cursor.execute("SELECT VERSION()")

        if 'mariadb' in self.cursor.fetchone()[0].lower():
            import ansible_collections.ansible.mysql.plugins.module_utils.implementations.mariadb.role as role_impl
        else:
            import ansible_collections.ansible.mysql.plugins.module_utils.implementations.mysql.role as role_impl

        return role_impl

    def check_users_in_db(self, users):
        """Check if users exist in a database.

        Args:
            users (list): List of tuples (username, hostname) to check.
        """
        for user in users:
            if user not in self.users:
                msg = 'User / role `%s` with host `%s` does not exist' % (user[0], user[1])
                self.module.fail_json(msg=msg)

    def filter_existing_users(self, users):
        for user in users:
            if user in self.users:
                yield user

    def __get_users(self):
        """Get users.

        Returns:
            list: List of tuples (username, hostname).
        """
        self.cursor.execute('SELECT User, Host FROM mysql.user')
        return self.cursor.fetchall()

    def get_users(self):
        """Get set of tuples (username, hostname) existing in a DB.

        Returns:
            self.users: Attribute value.
        """
        return self.users

    def get_grants(self, user, host):
        """Get grants.

        Args:
            user (str): User name
            host (str): Host name

        Returns:
            list: List of tuples like [(grant1,), (grant2,), ... ].
        """
        if host:
            self.cursor.execute('SHOW GRANTS FOR %s@%s', (user, host))
        else:
            self.cursor.execute('SHOW GRANTS FOR %s', (user,))

        return self.cursor.fetchall()


class MySQLQueryBuilder():
    """Class to build and return queries specific to MySQL.

    Args:
        name (str): Role name.
        host (str): Role host.

    Attributes:
        name (str): Role name.
        host (str): Role host.
    """
    def __init__(self, name, host):
        self.name = name
        self.host = host

    def role_exists(self):
        """Return a query to check if a role with self.name and self.host exists in a database.

        Returns:
            tuple: (query_string, tuple_containing_parameters).
        """
        return 'SELECT count(*) FROM mysql.user WHERE user = %s AND host = %s', (self.name, self.host)

    def role_grant(self, user):
        """Return a query to grant a role to a user or a role.

        Args:
            user (tuple): User / role to grant the role to in the form (username, hostname).

        Returns:
            tuple: (query_string, tuple_containing_parameters).
        """
        if user[1]:
            return 'GRANT %s@%s TO %s@%s', (self.name, self.host, user[0], user[1])
        else:
            return 'GRANT %s@%s TO %s', (self.name, self.host, user[0])

    def role_revoke(self, user):
        """Return a query to revoke a role from a user or role.

        Args:
            user (tuple): User / role to revoke the role from in the form (username, hostname).

        Returns:
            tuple: (query_string, tuple_containing_parameters).
        """
        if user[1]:
            return 'REVOKE %s@%s FROM %s@%s', (self.name, self.host, user[0], user[1])
        else:
            return 'REVOKE %s@%s FROM %s', (self.name, self.host, user[0])

    def role_create(self, admin=None):
        """Return a query to create a role.

        Args:
            admin (tuple): Admin user in the form (username, hostname).
                Because it is not supported by MySQL, we ignore it.

        Returns:
            tuple: (query_string, tuple_containing_parameters).
        """
        return 'CREATE ROLE %s', (self.name,)


class MariaDBQueryBuilder():
    """Class to build and return queries specific to MariaDB.

    Args:
        name (str): Role name.

    Attributes:
        name (str): Role name.
    """
    def __init__(self, name):
        self.name = name

    def role_exists(self):
        """Return a query to check if a role with self.name exists in a database.

        Returns:
            tuple: (query_string, tuple_containing_parameters).
        """
        return "SELECT count(*) FROM mysql.user WHERE user = %s AND is_role  = 'Y'", (self.name,)

    def role_grant(self, user):
        """Return a query to grant a role to a user or role.

        Args:
            user (tuple): User / role to grant the role to in the form (username, hostname).

        Returns:
            tuple: (query_string, tuple_containing_parameters).
        """
        if user[1]:
            return 'GRANT %s TO %s@%s', (self.name, user[0], user[1])
        else:
            return 'GRANT %s TO %s', (self.name, user[0])

    def role_revoke(self, user):
        """Return a query to revoke a role from a user or role.

        Args:
            user (tuple): User / role to revoke the role from in the form (username, hostname).

        Returns:
            tuple: (query_string, tuple_containing_parameters).
        """
        if user[1]:
            return 'REVOKE %s FROM %s@%s', (self.name, user[0], user[1])
        else:
            return 'REVOKE %s FROM %s', (self.name, user[0])

    def role_create(self, admin=None):
        """Return a query to create a role.

        Args:
            admin (tuple): Admin user in the form (username, hostname).

        Returns:
            tuple: (query_string, tuple_containing_parameters).
        """
        if not admin:
            return 'CREATE ROLE %s', (self.name,)

        if admin[1]:
            return 'CREATE ROLE %s WITH ADMIN %s@%s', (self.name, admin[0], admin[1])
        else:
            return 'CREATE ROLE %s WITH ADMIN %s', (self.name, admin[0])


class MySQLRoleImpl():
    """Class to work with MySQL role implementation.

    Args:
        module (AnsibleModule): Object of the AnsibleModule class.
        cursor (cursor): Cursor object of a database Python connector.
        name (str): Role name.
        host (str): Role host.

    Attributes:
        module (AnsibleModule): Object of the AnsibleModule class.
        cursor (cursor): Cursor object of a database Python connector.
        name (str): Role name.
        host (str): Role host.
    """
    def __init__(self, module, cursor, name, host):
        self.module = module
        self.cursor = cursor
        self.name = name
        self.host = host

    def set_default_role_all(self, user):
        """Run 'SET DEFAULT ROLE ALL TO' a user.

        Args:
            user (tuple): User / role to run the command against in the form (username, hostname).
        """
        if user[1]:
            self.cursor.execute('SET DEFAULT ROLE ALL TO %s@%s', (user[0], user[1]))
        else:
            self.cursor.execute('SET DEFAULT ROLE ALL TO %s', (user[0],))

    def get_admin(self):
        """Get a current admin of a role.

        Not supported by MySQL, so ignored here.
        """
        pass

    def set_admin(self, admin):
        """Set an admin of a role.

        Not supported by MySQL, so ignored here.

        TODO: Implement the feature if this gets supported.

        Args:
            admin (tuple): Admin user of the role in the form (username, hostname).
        """
        pass


class MariaDBRoleImpl():
    """Class to work with MariaDB role implementation.

    Args:
        module (AnsibleModule): Object of the AnsibleModule class.
        cursor (cursor): Cursor object of a database Python connector.
        name (str): Role name.

    Attributes:
        module (AnsibleModule): Object of the AnsibleModule class.
        cursor (cursor): Cursor object of a database Python connector.
        name (str): Role name.
    """
    def __init__(self, module, cursor, name):
        self.module = module
        self.cursor = cursor
        self.name = name

    def set_default_role_all(self, user):
        """Run 'SET DEFAULT ROLE ALL TO' a user.

        The command is not supported by MariaDB, ignored.

        Args:
            user (tuple): User / role to run the command against in the form (username, hostname).
        """
        pass

    def get_admin(self):
        """Get a current admin of a role.

        Returns:
            tuple: Of the form (username, hostname).
        """
        query = ("SELECT User, Host FROM mysql.roles_mapping "
                 "WHERE Role = %s and Admin_option = 'Y'")

        self.cursor.execute(query, (self.name,))
        return self.cursor.fetchone()

    def set_admin(self, admin):
        """Set an admin of a role.

        TODO: Implement changing when ALTER ROLE statement to
            change role's admin gets supported.

        Args:
            admin (tuple): Admin user of the role in the form (username, hostname).
        """
        admin_user = admin[0]
        admin_host = admin[1]
        current_admin = self.get_admin()

        if (admin_user, admin_host) != current_admin:
            msg = ('The "admin" option value and the current '
                   'roles admin (%s@%s) don not match. Ignored. '
                   'To change the admin, you need to drop and create the '
                   'role again.' % (current_admin[0], current_admin[1]))
            self.module.warn(msg)


class Role():
    """Class to work with MySQL role objects.

    Args:
        module (AnsibleModule): Object of the AnsibleModule class.
        cursor (cursor): Cursor object of a database Python connector.
        name (str): Role name.
        server (DbServer): Object of the DbServer class.

    Attributes:
        module (AnsibleModule): Object of the AnsibleModule class.
        cursor (cursor): Cursor object of a database Python connector.
        name (str): Role name.
        server (DbServer): Object of the DbServer class.
        host (str): Role's host.
        full_name (str): Role's full name.
        exists (bool): Indicates if a role exists or not.
        members (set): Set of current role's members.
    """
    def __init__(self, module, cursor, name, server):
        self.module = module
        self.cursor = cursor
        self.name = name
        self.server = server
        self.is_mariadb = self.server.is_mariadb()

        if self.is_mariadb:
            self.q_builder = MariaDBQueryBuilder(self.name)
            self.role_impl = MariaDBRoleImpl(self.module, self.cursor, self.name)
            self.full_name = '`%s`' % self.name
            self.host = ''
        else:
            self.host = '%'
            self.q_builder = MySQLQueryBuilder(self.name, self.host)
            self.role_impl = MySQLRoleImpl(self.module, self.cursor, self.name, self.host)
            self.full_name = '`%s`@`%s`' % (self.name, self.host)

        self.exists = self.__role_exists()
        self.members = set()

        if self.exists:
            self.members = self.__get_members()

    def __role_exists(self):
        """Check if a role exists.

        Returns:
            bool: True if the role exists, False if it does not.
        """
        self.cursor.execute(*self.q_builder.role_exists())
        return self.cursor.fetchone()[0] > 0

    def add(self, users, privs, check_mode=False, admin=False,
            set_default_role_all=True):
        """Add a role.

        Args:
            users (list): Role members.
            privs (str): String containing privileges.
            check_mode (bool): If True, just checks and does nothing.
            admin (tuple): Role's admin. Contains (username, hostname).
            set_default_role_all (bool): If True, runs SET DEFAULT ROLE ALL TO each member.

        Returns:
            bool: True if the state has changed, False if has not.
        """
        if check_mode:
            if not self.exists:
                return True
            return False

        self.cursor.execute(*self.q_builder.role_create(admin))

        if users:
            self.update_members(users, set_default_role_all=set_default_role_all)

        if privs:
            for db_table, priv in privs.items():
                privileges_grant(self.cursor, self.name, self.host,
                                 db_table, priv, tls_requires=None,
                                 maria_role=self.is_mariadb)

        return True

    def drop(self, check_mode=False):
        """Drop a role.

        Args:
            check_mode (bool): If True, just checks and does nothing.

        Returns:
            bool: True if the state has changed, False if has not.
        """
        if not self.exists:
            return False

        if check_mode and self.exists:
            return True

        self.cursor.execute('DROP ROLE %s', (self.name,))
        return True

    def update_members(self, users, check_mode=False, append_members=False,
                       set_default_role_all=True):
        """Add users to a role.

        Args:
            users (list): Role members.
            check_mode (bool): If True, just checks and does nothing.
            append_members (bool): If True, adds new members passed through users
                not touching current members.
            set_default_role_all (bool): If True, runs SET DEFAULT ROLE ALL TO each member.

        Returns:
            bool: True if the state has changed, False if has not.
        """
        if not users:
            return False

        changed = False
        for user in users:
            if user not in self.members:
                if check_mode:
                    return True

                self.cursor.execute(*self.q_builder.role_grant(user))

                if set_default_role_all:
                    self.role_impl.set_default_role_all(user)

                changed = True

        if append_members:
            return changed

        for user in self.members:
            if user not in users and user != ('root', 'localhost'):
                changed = self.__remove_member(user, check_mode)

        return changed

    def remove_members(self, users, check_mode=False):
        """Remove members from a role.

        Args:
            users (list): Role members.
            check_mode (bool): If True, just checks and does nothing.

        Returns:
            bool: True if the state has changed, False if has not.
        """
        if not users:
            return False

        changed = False
        for user in users:
            if user in self.members:
                changed = self.__remove_member(user, check_mode)

        return changed

    def __remove_member(self, user, check_mode=False):
        """Remove a member from a role.

        Args:
            user (str): Role member to remove.
            check_mode (bool): If True, just returns True and does nothing.

        Returns:
            bool: True if the state has changed, False if has not.
        """
        if check_mode:
            return True

        self.cursor.execute(*self.q_builder.role_revoke(user))

        return True

    def update(self, users, privs, check_mode=False,
               append_privs=False, subtract_privs=False,
               append_members=False, detach_members=False,
               admin=False, set_default_role_all=True):
        """Update a role.

        Update a role if needed.

        Todo: Implement changing of role's admin when ALTER ROLE statement
            to do that gets supported.

        Args:
            users (list): Role members.
            privs (str): String containing privileges.
            check_mode (bool): If True, just checks and does nothing.
            append_privs (bool): If True, adds new privileges passed through privs
                not touching current privileges.
            subtract_privs (bool): If True, revoke the privileges passed through privs
                not touching other existing privileges.
            append_members (bool): If True, adds new members passed through users
                not touching current members.
            detach_members (bool): If True, removes members passed through users from a role.
            admin (tuple): Role's admin. Contains (username, hostname).
            set_default_role_all (bool): If True, runs SET DEFAULT ROLE ALL TO each member.

        Returns:
            bool: True if the state has changed, False if has not.
        """
        changed = False
        members_changed = False

        if users:
            if detach_members:
                members_changed = self.remove_members(users, check_mode=check_mode)

            else:
                members_changed = self.update_members(users, check_mode=check_mode,
                                                      append_members=append_members,
                                                      set_default_role_all=set_default_role_all)

        if privs:
            result = user_mod(cursor=self.cursor, user=self.name, host=self.host,
                              host_all=None, password=None, encrypted=None, plugin=None,
                              plugin_auth_string=None, plugin_hash_string=None, salt=None,
                              new_priv=privs, append_privs=append_privs, subtract_privs=subtract_privs,
                              attributes=None, tls_requires=None, module=self.module, password_expire=None,
                              password_expire_interval=None, role=True, maria_role=self.is_mariadb)
            changed = result['changed']

        if admin:
            self.role_impl.set_admin(admin)

        changed = changed or members_changed

        return changed

    def __get_members(self):
        """Get current role's members.

        Returns:
            set: Members.
        """
        if self.is_mariadb:
            self.cursor.execute('select user, host from mysql.roles_mapping where role = %s', (self.name,))
        else:
            self.cursor.execute('select TO_USER as user, TO_HOST as host from mysql.role_edges where FROM_USER = %s', (self.name,))
        return set(self.cursor.fetchall())


def main():
    argument_spec = mysql_common_argument_spec()
    argument_spec.update(
        name=dict(type='str', required=True),
        state=dict(type='str', default='present', choices=['absent', 'present']),
        admin=dict(type='str'),
        priv=dict(type='raw'),
        append_privs=dict(type='bool', default=False),
        subtract_privs=dict(type='bool', default=False),
        members=dict(type='list', elements='str'),
        append_members=dict(type='bool', default=False),
        detach_members=dict(type='bool', default=False),
        check_implicit_admin=dict(type='bool', default=False),
        set_default_role_all=dict(type='bool', default=True),
        members_must_exist=dict(type='bool', default=True),
        column_case_sensitive=dict(type='bool', default=True),
        sql_log_bin=dict(type='bool', default=True),
    )
    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        mutually_exclusive=(
            ('append_members', 'detach_members'),
            ('admin', 'members'),
            ('admin', 'append_members'),
            ('admin', 'detach_members'),
            ('append_privs', 'subtract_privs'),
        ),
    )

    login_user = module.params['login_user']
    login_password = module.params['login_password']
    name = module.params['name']
    state = module.params['state']
    admin = module.params['admin']
    priv = module.params['priv']
    check_implicit_admin = module.params['check_implicit_admin']
    connect_timeout = module.params['connect_timeout']
    config_file = module.params['config_file']
    append_privs = module.params['append_privs']
    subtract_privs = module.boolean(module.params['subtract_privs'])
    members = module.params['members']
    append_members = module.params['append_members']
    detach_members = module.params['detach_members']
    ssl_cert = module.params['client_cert']
    ssl_key = module.params['client_key']
    ssl_ca = module.params['ca_cert']
    check_hostname = module.params['check_hostname']
    db = ''
    set_default_role_all = module.params['set_default_role_all']
    members_must_exist = module.params['members_must_exist']
    column_case_sensitive = module.params['column_case_sensitive']
    sql_log_bin = module.params['sql_log_bin']

    if priv and not isinstance(priv, (str, dict)):
        msg = ('The "priv" parameter must be str or dict '
               'but %s was passed' % type(priv))
        module.fail_json(msg=msg)

    if priv and isinstance(priv, dict):
        priv = convert_priv_dict_to_str(priv)

    if mysql_driver is None:
        module.fail_json(msg=mysql_driver_fail_msg)

    cursor = None
    try:
        if check_implicit_admin:
            try:
                cursor, db_conn = mysql_connect(module, 'root', '', config_file,
                                                ssl_cert, ssl_key, ssl_ca, db,
                                                connect_timeout=connect_timeout,
                                                check_hostname=check_hostname,
                                                autocommit=True)
            except Exception:
                pass

        if not cursor:
            cursor, db_conn = mysql_connect(module, login_user, login_password,
                                            config_file, ssl_cert, ssl_key,
                                            ssl_ca, db, connect_timeout=connect_timeout,
                                            check_hostname=check_hostname,
                                            autocommit=True)

    except Exception as e:
        module.fail_json(msg='unable to connect to database, '
                             'check login_user and login_password '
                             'are correct or %s has the credentials. '
                             'Exception message: %s' % (config_file, to_native(e)))

    if not sql_log_bin:
        cursor.execute("SET SQL_LOG_BIN=0;")

    # Set defaults
    changed = False

    impl = get_user_implementation(cursor)

    if priv is not None:
        try:
            mode = get_mode(cursor)
        except Exception as e:
            module.fail_json(msg=to_native(e))

        try:
            priv = privileges_unpack(priv, mode, column_case_sensitive, ensure_usage=not subtract_privs)
        except Exception as e:
            module.fail_json(msg='Invalid privileges string: %s' % to_native(e))

    server = DbServer(module, cursor)

    # Check if the server supports roles
    if not server.supports_roles():
        msg = ('Roles are not supported by the server. '
               'Minimal versions are MySQL 8.0.0 or MariaDB 10.0.5.')
        module.fail_json(msg=msg)

    if admin:
        if not server.is_mariadb():
            module.fail_json(msg='The "admin" option can be used only with MariaDB.')

        admin = normalize_users(module, [admin])[0]
        server.check_users_in_db([admin])

    if members:
        members = normalize_users(module, members, server.is_mariadb())
        if members_must_exist:
            server.check_users_in_db(members)
        else:
            members = list(server.filter_existing_users(members))

    # Main job starts here
    role = Role(module, cursor, name, server)

    try:
        if state == 'present':
            if not role.exists:
                if subtract_privs:
                    priv = None  # avoid granting unwanted privileges
                if detach_members:
                    members = None  # avoid adding unwanted members
                changed = role.add(members, priv, module.check_mode, admin,
                                   set_default_role_all)

            else:
                changed = role.update(members, priv, module.check_mode, append_privs, subtract_privs,
                                      append_members, detach_members, admin,
                                      set_default_role_all)

        elif state == 'absent':
            changed = role.drop(module.check_mode)

    except Exception as e:
        module.fail_json(msg=to_native(e))

    module.exit_json(changed=changed)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: networks_appliance_traffic_shaping_uplink_selection
short_description: Resource module for networks _appliance _traffic _shaping _uplink
  _selection
description:
  - Manage operation update of the resource networks _appliance _traffic _shaping
    _uplink _selection.
  - Update uplink selection settings for an MX network.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module
author: Francisco Munoz (@fmunoz)
options:
  activeActiveAutoVpnEnabled:
    description: Toggle for enabling or disabling active-active AutoVPN.
    type: bool
  defaultUplink:
    description: The default uplink. Must be a WAN interface 'wanX'.
    type: str
  failoverAndFailback:
    description: WAN failover and failback behavior.
    suboptions:
      immediate:
        description: Immediate WAN transition terminates all flows (new and existing)
          on current WAN when it is deemed unreliable.
        suboptions:
          enabled:
            description: Toggle for enabling or disabling immediate WAN failover and
              failback.
            type: bool
        type: dict
    type: dict
  loadBalancingEnabled:
    description: Toggle for enabling or disabling load balancing.
    type: bool
  networkId:
    description: NetworkId path parameter. Network ID.
    type: str
  vpnTrafficUplinkPreferences:
    description: Array of uplink preference rules for VPN traffic.
    elements: dict
    suboptions:
      failOverCriterion:
        description: Fail over criterion for this uplink preference rule. Must be
          one of 'poorPerformance' or 'uplinkDown'.
        type: str
      performanceClass:
        description: Performance class setting for this uplink preference rule.
        suboptions:
          builtinPerformanceClassName:
            description: Name of builtin performance class, must be present when performanceClass
              type is 'builtin', and value must be one of 'VoIP'.
            type: str
          customPerformanceClassId:
            description: ID of created custom performance class, must be present when
              performanceClass type is 'custom'.
            type: str
          type:
            description: Type of this performance class. Must be one of 'builtin'
              or 'custom'.
            type: str
        type: dict
      preferredUplink:
        description: Preferred uplink for this uplink preference rule. Must be one
          of 'wan1', 'wan2', 'bestForVoIP', 'loadBalancing', 'defaultUplink' or 'cellular',
          or any other valid uplink(wanX) if it applies to the network.
        type: str
      trafficFilters:
        description: Array of traffic filters for this uplink preference rule.
        elements: dict
        suboptions:
          type:
            description: Type of this traffic filter. Must be one of 'applicationCategory',
              'application' or 'custom'.
            type: str
          value:
            description: Value object of this traffic filter.
            suboptions:
              destination:
                description: Destination of this custom type traffic filter.
                suboptions:
                  cidr:
                    description: CIDR format address, or "any". E.g. "192.168.10.0/24",
                      "192.168.10.1" (same as "192.168.10.1/32"), "0.0.0.0/0" (same
                      as "any").
                    type: str
                  fqdn:
                    description: FQDN format address. Currently only availabe in 'destination'
                      of 'vpnTrafficUplinkPreference' object. E.g. 'www.google.com'.
                    type: str
                  host:
                    description: Host ID in the VLAN, should be used along with 'vlan',
                      and not exceed the vlan subnet capacity. Currently only available
                      under a template network.
                    type: int
                  network:
                    description: Meraki network ID. Currently only available under
                      a template network, and the value should be ID of either same
                      template network, or another template network currently. E.g.
                      "L_12345678".
                    type: str
                  port:
                    description: E.g. "any", "0" (also means "any"), "8080", "1-1024".
                    type: str
                  vlan:
                    description: VLAN ID of the configured VLAN in the Meraki network.
                      Currently only available under a template network.
                    type: int
                type: dict
              id:
                description: ID of this applicationCategory or application type traffic
                  filter. E.g. "meraki layer7/category/1", "meraki layer7/application/4".
                type: str
              protocol:
                description: Protocol of this custom type traffic filter. Must be
                  one of 'tcp', 'udp', 'icmp', 'icmp6' or 'any'.
                type: str
              source:
                description: Source of this custom type traffic filter.
                suboptions:
                  cidr:
                    description: CIDR format address, or "any". E.g. "192.168.10.0/24",
                      "192.168.10.1" (same as "192.168.10.1/32"), "0.0.0.0/0" (same
                      as "any").
                    type: str
                  host:
                    description: Host ID in the VLAN, should be used along with 'vlan',
                      and not exceed the vlan subnet capacity. Currently only available
                      under a template network.
                    type: int
                  network:
                    description: Meraki network ID. Currently only available under
                      a template network, and the value should be ID of either same
                      template network, or another template network currently. E.g.
                      "L_12345678".
                    type: str
                  port:
                    description: E.g. "any", "0" (also means "any"), "8080", "1-1024".
                    type: str
                  vlan:
                    description: VLAN ID of the configured VLAN in the Meraki network.
                      Currently only available under a template network.
                    type: int
                type: dict
            type: dict
        type: list
    type: list
  wanTrafficUplinkPreferences:
    description: Array of uplink preference rules for WAN traffic.
    elements: dict
    suboptions:
      preferredUplink:
        description: Preferred uplink for this uplink preference rule. Must be one
          of 'wan1' or 'wan2', or any other valid uplink(wanX) if it applies to the
          network.
        type: str
      trafficFilters:
        description: Array of traffic filters for this uplink preference rule.
        elements: dict
        suboptions:
          type:
            description: Type of this traffic filter. Must be one of 'custom'.
            type: str
          value:
            description: Value object of this traffic filter.
            suboptions:
              destination:
                description: Destination of this custom type traffic filter.
                suboptions:
                  cidr:
                    description: CIDR format address, or "any". E.g. "192.168.10.0/24",
                      "192.168.10.1" (same as "192.168.10.1/32"), "0.0.0.0/0" (same
                      as "any").
                    type: str
                  port:
                    description: E.g. "any", "0" (also means "any"), "8080", "1-1024".
                    type: str
                type: dict
              protocol:
                description: Protocol of this custom type traffic filter. Must be
                  one of 'tcp', 'udp', 'icmp6' or 'any'.
                type: str
              source:
                description: Source of this custom type traffic filter.
                suboptions:
                  cidr:
                    description: CIDR format address, or "any". E.g. "192.168.10.0/24",
                      "192.168.10.1" (same as "192.168.10.1/32"), "0.0.0.0/0" (same
                      as "any").
                    type: str
                  host:
                    description: Host ID in the VLAN, should be used along with 'vlan',
                      and not exceed the vlan subnet capacity. Currently only available
                      under a template network.
                    type: int
                  port:
                    description: E.g. "any", "0" (also means "any"), "8080", "1-1024".
                    type: str
                  vlan:
                    description: VLAN ID of the configured VLAN in the Meraki network.
                      Currently only available under a template network.
                    type: int
                type: dict
            type: dict
        type: list
    type: list
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for appliance updateNetworkApplianceTrafficShapingUplinkSelection
    description: Complete reference of the updateNetworkApplianceTrafficShapingUplinkSelection
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!update-network-appliance-traffic-shaping-uplink-selection
notes:
  - SDK Method used are
    appliance.Appliance.update_network_appliance_traffic_shaping_uplink_selection,
  - Paths used are
    put /networks/{networkId}/appliance/trafficShaping/uplinkSelection,
"""

EXAMPLES = r"""
- name: Update all
  cisco.meraki.networks_appliance_traffic_shaping_uplink_selection:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    activeActiveAutoVpnEnabled: true
    defaultUplink: wan1
    failoverAndFailback:
      immediate:
        enabled: true
    loadBalancingEnabled: true
    networkId: string
    vpnTrafficUplinkPreferences:
      - failOverCriterion: poorPerformance
        performanceClass:
          builtinPerformanceClassName: VoIP
          customPerformanceClassId: '123456'
          type: custom
        preferredUplink: bestForVoIP
        trafficFilters:
          - type: applicationCategory
            value:
              destination:
                cidr: any
                fqdn: www.google.com
                host: 254
                network: L_12345678
                port: 1-1024
                vlan: 10
              id: meraki:layer7/category/1
              protocol: tcp
              source:
                cidr: 192.168.1.0/24
                host: 200
                network: L_23456789
                port: any
                vlan: 20
    wanTrafficUplinkPreferences:
      - preferredUplink: wan1
        trafficFilters:
          - type: custom
            value:
              destination:
                cidr: any
                port: any
              protocol: tcp
              source:
                cidr: 192.168.1.0/24
                host: 254
                port: 1-1024
                vlan: 10
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "activeActiveAutoVpnEnabled": true,
      "defaultUplink": "string",
      "loadBalancingEnabled": true,
      "failoverAndFailback": {
        "immediate": {
          "enabled": true
        }
      },
      "wanTrafficUplinkPreferences": [
        {
          "trafficFilters": [
            {
              "type": "string",
              "value": {
                "protocol": "string",
                "source": {
                  "port": "string",
                  "cidr": "string",
                  "vlan": 0,
                  "host": 0
                },
                "destination": {
                  "port": "string",
                  "cidr": "string",
                  "applications": [
                    {
                      "id": "string",
                      "name": "string",
                      "type": "string"
                    }
                  ]
                }
              }
            }
          ],
          "preferredUplink": "string"
        }
      ],
      "vpnTrafficUplinkPreferences": [
        {
          "trafficFilters": [
            {
              "type": "string",
              "value": {
                "id": "string",
                "protocol": "string",
                "source": {
                  "port": "string",
                  "cidr": "string",
                  "network": "string",
                  "vlan": 0,
                  "host": 0
                },
                "destination": {
                  "port": "string",
                  "cidr": "string",
                  "network": "string",
                  "vlan": 0,
                  "host": 0,
                  "fqdn": "string"
                }
              }
            }
          ],
          "preferredUplink": "string",
          "failOverCriterion": "string",
          "performanceClass": {
            "type": "string",
            "builtinPerformanceClassName": "string",
            "customPerformanceClassId": "string"
          }
        }
      ]
    }
"""



#!/usr/bin/python
#
# Copyright (c) 2022 xuzhang3 (@xuzhang3), Fred-sun (@Fred-sun)
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type


DOCUMENTATION = '''
---
module: azure_rm_bastionhost

version_added: "1.13.0"

short_description: Managed bastion host resource

description:
    - Create, Update or Delete bastion host resource.

options:
    name:
        description:
            - The name of the bastion host.
        type: str
        required: True
    resource_group:
        description:
            - The name of the resource group.
        type: str
        required: True
    location:
        description:
            - The resource location.
        type: str
    sku:
        description:
            - The sku of this Bastion Host.
        type: dict
        suboptions:
            name:
                description:
                    - The name of the SKU.
                type: str
                choices:
                    - Standard
                    - Basic
    enable_tunneling:
        description:
            - Enable or Disable Tunneling feature of the Bastion Host resource.
        type: bool
    enable_shareable_link:
        description:
            - Enable or Disable Shareable Link of the Bastion Host resource.
        type: bool
    enable_ip_connect:
        description:
            - Enable or Disable IP Connect feature of the Bastion Host resource.
        type: bool
    enable_file_copy:
        description:
            - Enable or Disable File Copy feature of the Bastion Host resource.
        type: bool
    scale_units:
        description:
            - The scale units for the Bastion Host resource.
        type: int
    disable_copy_paste:
        description:
            - Enable or Disable Copy or Paste feature of the Bastion Host resource.
        type: bool
    ip_configurations:
        description:
            - An array of bastion host IP configurations.
        type: list
        elements: dict
        suboptions:
            name:
                description:
                    - The name of bastion host ip configuration.
                type: str
            subnet:
                description:
                    - Reference of the subnet resource.
                type: dict
                suboptions:
                    id:
                        description:
                            - The ID of the Subnet.
                        type: str
            public_ip_address:
                description:
                    - Reference of the PublicIP resource.
                type: dict
                suboptions:
                    id:
                        description:
                            - The ID of the public IP.
                        type: str
            private_ip_allocation_method:
                description:
                    - Private IP allocation method.
                type: str
                choices:
                    - Static
                    - Dynamic
    state:
        description:
            - Assert the state of the pirvate link service.
            - Use I(state=present) to create or update the link service and I(state=absent) to delete it.
        type: str
        default: present
        choices:
            - present
            - absent

extends_documentation_fragment:
    - azure.azcollection.azure
    - azure.azcollection.azure_tags

author:
    - xuzhang3 (@xuzhang3)
    - Fred-sun (@Fred-sun)
'''

EXAMPLES = '''
- name: Create bastion host info
  azure_rm_bastionhost:
    name: bastion-name
    resource_group: myResourceGroup
    ip_configurations:
      - name: testip_configuration
        subnet:
          id: "{{ subnet_output.state.id }}"
        public_ip_address:
          id: "{{ publicIP_output.publicipaddresses[0].id }}"
        private_ip_allocation_method: Dynamic
    sku:
      name: Standard
    enable_tunneling: false
    enable_shareable_link: false
    enable_ip_connect: false
    enable_file_copy: false
    scale_units: 6
    disable_copy_paste: false
    tags:
      key1: value1

- name: Create bastion host info
  azure_rm_bastionhost:
    name: bastion-name
    resource_group: myResourceGroup
    state: absent
'''

RETURN = '''
bastion_host:
    description:
        - List of Azure bastion host info.
    returned: always
    type: complex
    contains:
        id:
            description:
                - Resource ID of the Azure bastion host.
            sample: "/subscriptions/xxx-xxx/resourceGroups/myResourceGroup/providers/Microsoft.Network/bastionHosts/testbastion"
            returned: always
            type: str
        name:
            description:
                - Name of the Azure bastion host.
            returned: always
            type: str
            sample: linkservice
        location:
            description:
                - Resource location.
            returned: always
            type: str
            sample: eastus
        etag:
            description:
                - A unique read-only string that changes whenever the resource is updated.
            type: str
            returned: always
            sample: "fb0e3a90-6afa-4a01-9171-9c84d144a0f3"
        type:
            description:
                - The resource type.
            type: str
            returned: always
            sample: Microsoft.Network/bastionHosts
        tags:
            description:
                - The resource tags.
            type: list
            returned: always
            sample: { 'key1': 'value1' }
        provisioning_state:
            description:
                - The provisioning state of the bastion host resource.
            type: str
            returned: always
            sample: Succeeded
        scale_units:
            description:
                - The scale units for the Bastion Host resource.
            type: int
            returned: always
            sample: 2
        enable_tunneling:
            description:
                - Enable/Disable Tunneling feature of the Bastion Host resource.
            type: bool
            returned: always
            sample: False
        enable_shareable_link:
            description:
                - Enable/Disable Shareable Link of the Bastion Host resource.
            type: bool
            returned: always
            sample: False
        enable_ip_connect:
            description:
                - Enable/Disable IP Connect feature of the Bastion Host resource.
            type: bool
            returned: always
            sample: False
        enable_file_copy:
            description:
                - Enable/Disable File Copy feature of the Bastion Host resource.
            type: bool
            returned: always
            sample: False
        dns_name:
            description:
                - FQDN for the endpoint on which bastion host is accessible.
            type: str
            returned: always
            sample: bst-0ca1e1b6-9969-4167-be54-5972e1395c25.bastion.azure.com
        disable_copy_paste:
            description:
                - Enable/Disable Copy/Paste feature of the Bastion Host resource.
            type: bool
            returned: always
            sample: False
        sku:
            description:
                - The sku of this Bastion Host.
            type: complex
            returned: always
            contains:
                name:
                    description:
                        -  The name of this Bastion Host.
                    type: str
                    returned: always
                    sample: Standard
        ip_configurations:
            description:
                - An array of bastion host IP configurations.
            type: complex
            returned: always
            contains:
                name:
                    description:
                        - Name of the resource that is unique within a resource group.
                        - This name can be used to access the resource.
                    type: str
                    returned: always
                    sample: IpConf
                private_ip_allocation_method:
                    description:
                        - Private IP allocation method.
                    type: str
                    returned: always
                    sample: Static
                public_ip_address:
                    description:
                        - Reference of the PublicIP resource.
                    type: complex
                    returned: always
                    contains:
                        id:
                            description:
                                - The ID of the public IP address.
                            returned: always
                            type: str
                            sample: "/subscriptions/xxx-xxx/resourceGroups/myRG/providers/Microsoft.Network/publicIPAddresses/Myip"
                subnet:
                    description:
                        - The reference to the subnet resource.
                    returned: always
                    type: str
                    contains:
                        id:
                            description:
                                - The ID of the subnet..
                            returned: always
                            type: str
                            sample:  "/subscriptions/xxx-xxx/resourceGroups/myRG/providers/Microsoft.Network/virtualNetworks/vnet/subnets/AzureBastionSubnet"
'''

try:
    from azure.core.exceptions import ResourceNotFoundError
except Exception:
    # This is handled in azure_rm_common
    pass


from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase


sku_spec = dict(
    name=dict(type='str', choices=['Standard', 'Basic'])
)

subnet_spec = dict(
    id=dict(type='str')
)

public_ip_address_spec = dict(
    id=dict(type='str')
)


class AzureRMBastionHost(AzureRMModuleBase):

    def __init__(self):

        self.module_arg_spec = dict(
            name=dict(type="str", required=True),
            resource_group=dict(type="str", required=True),
            state=dict(type='str', default='present', choices=['present', 'absent']),
            location=dict(type='str'),
            ip_configurations=dict(
                type='list',
                elements='dict',
                options=dict(
                    name=dict(type='str'),
                    subnet=dict(type='dict', options=subnet_spec),
                    public_ip_address=dict(type='dict', options=public_ip_address_spec),
                    private_ip_allocation_method=dict(type='str', choices=['Static', 'Dynamic'])
                )
            ),
            sku=dict(type='dict', options=sku_spec),
            enable_tunneling=dict(type='bool'),
            enable_shareable_link=dict(type='bool'),
            enable_ip_connect=dict(type='bool'),
            enable_file_copy=dict(type='bool'),
            scale_units=dict(type='int'),
            disable_copy_paste=dict(type='bool')
        )

        self.name = None
        self.resource_group = None
        self.location = None
        self.tags = None
        self.state = None
        self.results = dict(
            changed=False,
        )
        self.body = {}

        super(AzureRMBastionHost, self).__init__(self.module_arg_spec,
                                                 supports_check_mode=True,
                                                 supports_tags=True,
                                                 facts_module=False)

    def exec_module(self, **kwargs):
        for key in list(self.module_arg_spec.keys()) + ['tags']:
            if hasattr(self, key):
                setattr(self, key, kwargs[key])
            elif kwargs[key] is not None:
                self.body[key] = kwargs[key]

        old_response = self.get_item()
        result = None
        changed = False

        if not self.location:
            resource_group = self.get_resource_group(self.resource_group)
            self.location = resource_group.location

        self.body['location'] = self.location
        self.body['tags'] = self.tags

        if self.state == 'present':
            if old_response:
                update_tags, tags = self.update_tags(old_response['tags'])
                if update_tags:
                    changed = True
                self.body['tags'] = tags

                if self.body.get('disable_copy_paste') is not None:
                    if bool(self.body.get('disable_copy_paste')) != bool(old_response['disable_copy_paste']):
                        changed = True
                else:
                    self.body['disable_copy_paste'] = old_response['disable_copy_paste']

                if self.body.get('enable_file_copy') is not None:
                    if bool(self.body.get('enable_file_copy')) != bool(old_response['enable_file_copy']):
                        changed = True
                else:
                    self.body['enable_file_copy'] = old_response['enable_file_copy']

                if self.body.get('enable_ip_connect') is not None:
                    if bool(self.body.get('enable_ip_connect')) != bool(old_response['enable_ip_connect']):
                        changed = True
                else:
                    self.body['enable_ip_connect'] = old_response['enable_ip_connect']

                if self.body.get('enable_shareable_link') is not None:
                    if bool(self.body.get('enable_shareable_link')) != bool(old_response['enable_shareable_link']):
                        changed = True
                else:
                    self.body['enable_shareable_link'] = old_response['enable_shareable_link']

                if self.body.get('enable_tunneling') is not None:
                    if bool(self.body.get('enable_tunneling')) != bool(old_response['enable_tunneling']):
                        changed = True
                else:
                    self.body['enable_tunneling'] = old_response['enable_tunneling']

                if self.body.get('scale_units') is not None:
                    if self.body.get('scale_units') != old_response['scale_units']:
                        changed = True
                else:
                    self.body['scale_units'] = old_response['scale_units']

                if self.body.get('sku') is not None:
                    if self.body.get('sku') != old_response['sku']:
                        changed = True
                else:
                    self.body['sku'] = old_response['sku']

                if self.body.get('ip_configurations') is not None:
                    if self.body['ip_configurations'] != old_response['ip_configurations']:
                        self.fail("Bastion Host IP configuration not support to update!")
                else:
                    self.body['ip_configurations'] = old_response['ip_configurations']
            else:
                changed = True

            if changed:
                if self.check_mode:
                    self.log("Check mode test. The bastion host is exist, will be create or updated")
                else:
                    result = self.create_or_update(self.body)
            else:
                if self.check_mode:
                    self.log("Check mode test. The Azure Bastion Host is exist, No operation in this task")
                else:
                    self.log("The Azure Bastion Host is exist, No operation in this task")
                    result = old_response
        else:
            if old_response:
                changed = True
                if self.check_mode:
                    self.log("Check mode test. The bastion host is exist, will be deleted")
                else:
                    result = self.delete_resource()
            else:
                if self.check_mode:
                    self.log("The bastion host isn't exist, no action")
                else:
                    self.log("The bastion host isn't exist, don't need to delete")

        self.results["bastion_host"] = result
        self.results['changed'] = changed
        return self.results

    def get_item(self):
        self.log("Get properties for {0} in {1}".format(self.name, self.resource_group))
        try:
            response = self.network_client.bastion_hosts.get(self.resource_group, self.name)
            return self.bastion_to_dict(response)
        except ResourceNotFoundError:
            self.log("Could not get info for {0} in {1}".format(self.name, self.resource_group))

        return []

    def create_or_update(self, parameters):
        self.log("Create or update the bastion host for {0} in {1}".format(self.name, self.resource_group))
        try:
            response = self.network_client.bastion_hosts.begin_create_or_update(self.resource_group, self.name, parameters)

            result = self.network_client.bastion_hosts.get(self.resource_group, self.name)
            return self.bastion_to_dict(result)
        except Exception as ec:
            self.fail("Create or Update {0} in {1} failed, mesage {2}".format(self.name, self.resource_group, ec))

        return []

    def delete_resource(self):
        self.log("delete the bastion host for {0} in {1}".format(self.name, self.resource_group))
        try:
            response = self.network_client.bastion_hosts.begin_delete(self.resource_group, self.name)
        except Exception as ec:
            self.fail("Delete {0} in {1} failed, message {2}".format(self.name, self.resource_group, ec))

        return []

    def bastion_to_dict(self, bastion_info):
        bastion = bastion_info.as_dict()
        result = dict(
            id=bastion.get("id"),
            name=bastion.get('name'),
            type=bastion.get('type'),
            etag=bastion.get('etag'),
            location=bastion.get('location'),
            tags=bastion.get('tags'),
            sku=dict(),
            ip_configurations=list(),
            dns_name=bastion.get('dns_name'),
            provisioning_state=bastion.get('provisioning_state'),
            scale_units=bastion.get('scale_units'),
            disable_copy_paste=bastion.get('disable_copy_paste'),
            enable_file_copy=bastion.get('enable_file_copy'),
            enable_ip_connect=bastion.get('enable_ip_connect'),
            enable_shareable_link=bastion.get('enable_tunneling'),
            enable_tunneling=bastion.get('enable_tunneling')
        )

        if bastion.get('sku'):
            result['sku']['name'] = bastion['sku']['name']

        if bastion.get('ip_configurations'):
            for items in bastion['ip_configurations']:
                result['ip_configurations'].append(
                    {
                        "name": items['name'],
                        "subnet": dict(id=items['subnet']['id']),
                        "public_ip_address": dict(id=items['public_ip_address']['id']),
                        "private_ip_allocation_method": items['private_ip_allocation_method'],
                    }
                )
        return result


def main():
    AzureRMBastionHost()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# (c) 2018, Ansible by Red Hat, inc
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import absolute_import, division, print_function


__metaclass__ = type


DOCUMENTATION = """
module: cli_backup
author: Katherine Case (@Qalthos)
short_description: Back up device configuration from network devices over network_cli
description:
- This module provides platform agnostic way of backing up text based configuration from
  network devices over network_cli connection plugin.
version_added: 4.2.0
extends_documentation_fragment:
- ansible.netcommon.network_agnostic
options:
  defaults:
    description:
    - The I(defaults) argument will influence how the running-config is collected
      from the device.  When the value is set to true, the command used to collect
      the running-config is append with the all keyword.  When the value is set to
      false, the command is issued without the all keyword.
    default: no
    type: bool
  filename:
    description:
    - The filename to be used to store the backup configuration. If the filename
      is not given it will be generated based on the hostname, current time and
      date in format defined by <hostname>_config.<current-date>@<current-time>
    type: str
  dir_path:
    description:
    - This option provides the path ending with directory name in which the backup
      configuration file will be stored. If the directory does not exist it will
      be first created and the filename is either the value of C(filename) or
      default filename as described in C(filename) options description. If the
      path value is not given in that case a I(backup) directory will be created
      in the current working directory and backup configuration will be copied
      in C(filename) within I(backup) directory.
    type: path
"""

EXAMPLES = """
- name: configurable backup path
  ansible.netcommon.cli_backup:
    filename: backup.cfg
    dir_path: /home/user
"""

RETURN = """
backup_path:
  description: The full path to the backup file
  returned: always
  type: str
  sample: /playbooks/ansible/backup/hostname_config.2016-07-16@22:28:34
"""

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.connection import Connection


def validate_args(module, device_operations):
    """validate param if it is supported on the platform"""
    feature_list = [
        "defaults",
    ]

    for feature in feature_list:
        if module.params[feature]:
            supports_feature = device_operations.get("supports_%s" % feature)
            if supports_feature is None:
                module.fail_json(
                    msg="This platform does not specify whether %s is supported or not. "
                    "Please report an issue against this platform's cliconf plugin." % feature
                )
            elif not supports_feature:
                module.fail_json(msg="Option %s is not supported on this platform" % feature)


def main():
    """main entry point for execution"""
    argument_spec = dict(
        defaults=dict(default=False, type="bool"),
        filename=dict(),
        dir_path=dict(type="path"),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
    )

    result = {"changed": False}

    connection = Connection(module._socket_path)
    capabilities = module.from_json(connection.get_capabilities())

    if capabilities:
        device_operations = capabilities.get("device_operations", dict())
        validate_args(module, device_operations)
    else:
        device_operations = dict()

    if module.params["defaults"]:
        if "get_default_flag" in capabilities.get("rpc"):
            flags = connection.get_default_flag()
        else:
            flags = "all"
    else:
        flags = []

    running = connection.get_config(flags=flags)
    result["__backup__"] = running

    module.exit_json(**result)


if __name__ == "__main__":
    main()



#!/usr/bin/python
#
# Copyright (c) 2025 Klaas Demter (@Klaas-)
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = '''
---
module: azure_rm_monitordatacollectionrules
version_added: "3.7.0"
short_description: Create, update and delete Data Collection Rules
description:
    - Create, update and delete Data Collection Rules

options:
    name:
        description:
            - The name of the data collection rule you're creating/changing
        required: true
        type: str
    resource_group:
        description:
            - The name of the resource group
        required: true
        type: str
    location:
        description:
            - Location of the data colelction rule
            - defaults to location of exiting data collection rule or
            - location of the resource group if unspecified
        required: false
        type: str
    kind:
        description:
            - Kind of the data collection rule
            - Use C(Linux) for Linux.
            - Use C(Windows) for Windows.
        type: str
        choices:
            - Linux
            - Windows
    description:
        description:
            - Description for the data collection rule
        type: str
    data_collection_endpoint_id:
        description:
            - The resource ID of the data collection endpoint that this rule can be used with
        type: str
    stream_declarations:
        description:
            - Declaration of a custom stream. Sub dict is a list of columns used by data in this stream.
            - top level key is the name of the stream_declaration
            - U(https://learn.microsoft.com/en-us/python/api/azure-mgmt-monitor/azure.mgmt.monitor.v2022_06_01.models.streamdeclaration?view=azure-python)
        type: dict
        suboptions:
            name_of_stream:
                description:
                    - Name of the stream
                type: dict
                suboptions:
                    columns:
                        description:
                            - Declaration of a custom stream.
                        type: list
                        elements: dict
                        suboptions:
                            name:
                                description:
                                    - The name of the column.
                                type: str
                            type:
                                description:
                                    - The type of the column data.
                                type: str
                                choices:
                                    - string
                                    - int
                                    - long
                                    - real
                                    - boolean
                                    - datetime
                                    - dynamic
    data_sources:
        description:
            - The specification of data sources.
            - This property is optional and can be omitted if the rule is meant to be used via direct calls to the provisioned endpoint.
            - >-
              U(https://learn.microsoft.com/en-us/python/api/azure-mgmt-monitor/azure.mgmt.monitor.v2022_06_01.models.datacollectionruledatasources?view=azure-python)
        type: dict
        suboptions:
            performance_counters:
                description:
                    - Definition of which performance counters will be collected and how they will be collected by this data collection rule.
                    - Collected from both Windows and Linux machines where the counter is present.
                    - >-
                      U(https://learn.microsoft.com/en-us/python/api/azure-mgmt-monitor/azure.mgmt.monitor.v2022_06_01.models.perfcounterdatasource?view=azure-python)
                type: list
                elements: dict
                suboptions:
                    streams:
                        description:
                            - List of streams that this data source will be sent to.
                            - A stream indicates what schema will be used for this data and usually what table in Log Analytics the data will be sent to.
                        type: list
                        elements: str
                    sampling_frequency_in_seconds:
                        description:
                            - The number of seconds between consecutive counter measurements (samples).
                        type: int
                    counter_specifiers:
                        description:
                            - A list of specifier names of the performance counters you want to collect.
                            - Use a wildcard (*) to collect a counter for all instances.
                            - To get a list of performance counters on Windows, run the command 'typeperf'
                        type: list
                        elements: str
                    name:
                        description:
                            - A friendly name for the data source.
                            - This name should be unique across all data sources (regardless of type) within the data collection rule.
                        type: str
            windows_event_logs:
                description:
                    - Definition of which Windows Event Log events will be collected and how they will be collected.
                    - Only collected from Windows machines.
                    - >-
                      U(https://learn.microsoft.com/en-us/python/api/azure-mgmt-monitor/azure.mgmt.monitor.v2022_06_01.models.windowseventlogdatasource?view=azure-python)
                type: list
                elements: dict
                suboptions:
                    streams:
                        description:
                            - List of streams that this data source will be sent to.
                            - A stream indicates what schema will be used for this data and usually what table in Log Analytics the data will be sent to.
                        type: list
                        elements: str
                    x_path_queries:
                        description:
                            - A list of Windows Event Log queries in XPATH format.
                        type: list
                        elements: str
                    name:
                        description:
                            - A friendly name for the data source.
                            - This name should be unique across all data sources (regardless of type) within the data collection rule.
                        type: str
            syslog:
                description:
                    - Definition of which syslog data will be collected and how it will be collected. Only collected from Linux machines.
                    - >-
                      U(https://learn.microsoft.com/en-us/python/api/azure-mgmt-monitor/azure.mgmt.monitor.v2022_06_01.models.syslogdatasource?view=azure-python)
                type: list
                elements: dict
                suboptions:
                    streams:
                        description:
                            - List of streams that this data source will be sent to.
                            - A stream indicates what schema will be used for this data and usually what table in Log Analytics the data will be sent to.
                        type: list
                        elements: str
                    facility_names:
                        description:
                            - The list of facility names.
                        type: list
                        elements: str
                    log_levels:
                        description:
                            - The log levels to collect.
                        type: list
                        elements: str
                    name:
                        description:
                            - A friendly name for the data source.
                            - This name should be unique across all data sources (regardless of type) within the data collection rule.
                        type: str
            extensions:
                description:
                    - Definition of which data will be collected from a separate VM extension that integrates with the Azure Monitor Agent.
                    - Collected from either Windows and Linux machines, depending on which extension is defined.
                    - >-
                      U(https://learn.microsoft.com/en-us/python/api/azure-mgmt-monitor/azure.mgmt.monitor.v2022_06_01.models.extensiondatasource?view=azure-python)
                type: list
                elements: dict
                suboptions:
                    streams:
                        description:
                            - List of streams that this data source will be sent to.
                            - A stream indicates what schema will be used for this data and usually what table in Log Analytics the data will be sent to.
                        type: list
                        elements: str
                    extension_name:
                        description:
                            - The name of the VM extension.
                        type: list
                        elements: str
                    extension_settings:
                        description:
                            - The extension settings. The format is specific for particular extension.
                        type: str
                    input_data_sources:
                        description:
                            - The list of data sources this extension needs data from.
                        type: list
                        elements: str
                    name:
                        description:
                            - A friendly name for the data source.
                            - This name should be unique across all data sources (regardless of type) within the data collection rule.
                        type: str
            log_files:
                description:
                    - Definition of which custom log files will be collected by this data collection rule.
                    - >-
                      U(https://learn.microsoft.com/en-us/python/api/azure-mgmt-monitor/azure.mgmt.monitor.v2022_06_01.models.logfilesdatasource?view=azure-python)
                type: list
                elements: dict
                suboptions:
                    streams:
                        description:
                            - List of streams that this data source will be sent to.
                            - A stream indicates what schema will be used for this data and usually what table in Log Analytics the data will be sent to.
                        type: list
                        elements: str
                    file_patterns:
                        description:
                            - File Patterns where the log files are located
                        type: list
                        elements: str
                    format:
                        description:
                            - The data format of the log files.
                        type: str
                    settings:
                        description:
                            - The log files specific settings.
                        type: dict
                        suboptions:
                            text:
                                description:
                                    - Text settings
                                type: str
                    name:
                        description:
                            - A friendly name for the data source.
                            - This name should be unique across all data sources (regardless of type) within the data collection rule.
                        type: str
            iis_logs:
                description:
                    - Enables IIS logs to be collected by this data collection rule.
                    - >-
                      U(https://learn.microsoft.com/en-us/python/api/azure-mgmt-monitor/azure.mgmt.monitor.v2022_06_01.models.iislogsdatasource?view=azure-python)
                type: list
                elements: dict
                suboptions:
                    streams:
                        description:
                            - IIS streams.
                        type: list
                        elements: str
                    log_directories:
                        description:
                            - Absolute paths file location.
                        type: list
                        elements: str
                    name:
                        description:
                            - A friendly name for the data source.
                            - This name should be unique across all data sources (regardless of type) within the data collection rule.
                        type: str
            windows_firewall_logs:
                description:
                    - Enables Firewall logs to be collected by this data collection rule.
                    - >-
                      U(https://learn.microsoft.com/en-us/python/api/azure-mgmt-monitor/azure.mgmt.monitor.v2022_06_01.models.windowsfirewalllogsdatasource?view=azure-python)
                type: list
                elements: dict
                suboptions:
                    streams:
                        description:
                            - Firewall logs streams.
                        type: list
                        elements: str
                    name:
                        description:
                            - A friendly name for the data source.
                            - This name should be unique across all data sources (regardless of type) within the data collection rule.
                        type: str
            prometheus_forwarder:
                description:
                    - Definition of Prometheus metrics forwarding configuration.
                    - >-
                      U(https://learn.microsoft.com/en-us/python/api/azure-mgmt-monitor/azure.mgmt.monitor.v2022_06_01.models.prometheusforwarderdatasource?view=azure-python)
                type: list
                elements: dict
                suboptions:
                    streams:
                        description:
                            - List of streams that this data source will be sent to.
                        type: list
                        elements: str
                    label_include_filter:
                        description:
                            - The list of label inclusion filters in the form of label "name-value" pairs.
                            - Currently only one label is supported "microsoft_metrics_include_label". Label values are matched case-insensitively.
                        type: dict
                    name:
                        description:
                            - A friendly name for the data source.
                            - This name should be unique across all data sources (regardless of type) within the data collection rule.
                        type: str
            platform_telemetry:
                description:
                    - Definition of platform telemetry data source configuration.
                    - >-
                      U(https://learn.microsoft.com/en-us/python/api/azure-mgmt-monitor/azure.mgmt.monitor.v2022_06_01.models.platformtelemetrydatasource?view=azure-python)
                type: list
                elements: dict
                suboptions:
                    streams:
                        description:
                            - List of platform telemetry streams to collect.
                        type: list
                        elements: str
                    name:
                        description:
                            - A friendly name for the data source.
                            - This name should be unique across all data sources (regardless of type) within the data collection rule.
                        type: str
            data_imports:
                description:
                    - Specifications of pull based data sources.
                    - >-
                      U(https://learn.microsoft.com/en-us/python/api/azure-mgmt-monitor/azure.mgmt.monitor.v2022_06_01.models.datasourcesspecdataimports?view=azure-python)
                type: dict
                suboptions:
                    event_hub:
                        description:
                            - Definition of Event Hub configuration.
                        type: dict
                        suboptions:
                            name:
                                description:
                                    - A friendly name for the data source.
                                    - This name should be unique across all data sources (regardless of type) within the data collection rule.
                                type: str
                            consumer_group:
                                description:
                                    - Event Hub consumer group name.
                                type: str
                            stream:
                                description:
                                    - The stream to collect from EventHub.
                                type: str
    destinations:
        description:
            - >-
              U(https://learn.microsoft.com/en-us/python/api/azure-mgmt-monitor/azure.mgmt.monitor.v2022_06_01.models.datacollectionruledestinations?view=azure-python)
        type: dict
        suboptions:
            log_analytics:
                description:
                    - List of Log Analytics destinations.
                    - >-
                      U(https://learn.microsoft.com/en-us/python/api/azure-mgmt-monitor/azure.mgmt.monitor.v2022_06_01.models.loganalyticsdestination?view=azure-python)
                type: list
                elements: dict
                suboptions:
                    workspace_resource_id:
                        description:
                            - The resource ID of the Log Analytics workspace.
                        type: str
                    name:
                        description:
                            - A friendly name for the destination.
                            - This name should be unique across all destinations (regardless of type) within the data collection rule.
                        type: str
            monitoring_accounts:
                description:
                    - List of monitoring account destinations.
                    - >-
                      U(https://learn.microsoft.com/en-us/python/api/azure-mgmt-monitor/azure.mgmt.monitor.v2022_06_01.models.monitoringaccountdestination?view=azure-python)
                type: list
                elements: dict
                suboptions:
                    account_resource_id:
                        description:
                            - The resource ID of the monitoring account.
                        type: str
                    name:
                        description:
                            - A friendly name for the destination.
                            - This name should be unique across all destinations (regardless of type) within the data collection rule.
                        type: str
            azure_monitor_metrics:
                description:
                    - Azure Monitor Metrics destination
                    - >-
                      U(https://learn.microsoft.com/en-us/python/api/azure-mgmt-monitor/azure.mgmt.monitor.v2022_06_01.models.destinationsspecazuremonitormetrics?view=azure-python)
                type: str
            event_hubs:
                description:
                    - List of Event Hubs destinations.
                    - >-
                      U(https://learn.microsoft.com/en-us/python/api/azure-mgmt-monitor/azure.mgmt.monitor.v2022_06_01.models.eventhubdestination?view=azure-python)
                type: list
                elements: dict
                suboptions:
                    event_hub_resource_id:
                        description:
                            - The resource ID of the event hub.
                        type: str
                    name:
                        description:
                            - A friendly name for the destination.
                            - This name should be unique across all destinations (regardless of type) within the data collection rule.
                        type: str
            event_hubs_direct:
                description:
                    - List of Event Hubs Direct destinations.
                    - >-
                      U(https://learn.microsoft.com/en-us/python/api/azure-mgmt-monitor/azure.mgmt.monitor.v2022_06_01.models.eventhubdirectdestination?view=azure-python)
                type: list
                elements: dict
                suboptions:
                    event_hub_resource_id:
                        description:
                            - The resource ID of the event hub.
                        type: str
                    name:
                        description:
                            - A friendly name for the destination.
                            - This name should be unique across all destinations (regardless of type) within the data collection rule.
                        type: str
            storage_blobs_direct:
                description:
                    - List of Storage Blob Direct destinations. To be used only for sending data directly to store from the agent.
                    - >-
                      U(https://learn.microsoft.com/en-us/python/api/azure-mgmt-monitor/azure.mgmt.monitor.v2022_06_01.models.storageblobdestination?view=azure-python)
                type: list
                elements: dict
                suboptions:
                    container_name:
                        description:
                            - The container name of the Storage Blob.
                        type: str
                    storage_account_resource_id:
                        description:
                            - The resource ID of the storage account.
                        type: str
                    name:
                        description:
                            - A friendly name for the destination.
                            - This name should be unique across all destinations (regardless of type) within the data collection rule.
                        type: str
            storage_tables_direct:
                description:
                    - List of Storage Table Direct destinations.
                    - >-
                      U(https://learn.microsoft.com/en-us/python/api/azure-mgmt-monitor/azure.mgmt.monitor.v2022_06_01.models.storagetabledestination?view=azure-python)
                type: list
                elements: dict
                suboptions:
                    table_name:
                        description:
                            - The name of the Storage Table.
                        type: str
                    storage_account_resource_id:
                        description:
                            - The resource ID of the storage account.
                        type: str
                    name:
                        description:
                            - A friendly name for the destination.
                            - This name should be unique across all destinations (regardless of type) within the data collection rule.
                        type: str
            storage_accounts:
                description:
                    - List of storage accounts destinations.
                    - >-
                      U(https://learn.microsoft.com/en-us/python/api/azure-mgmt-monitor/azure.mgmt.monitor.v2022_06_01.models.storageblobdestination?view=azure-python)
                type: list
                elements: dict
                suboptions:
                    container_name:
                        description:
                            - The container name of the Storage Blob.
                        type: str
                    storage_account_resource_id:
                        description:
                            - The resource ID of the storage account.
                        type: str
                    name:
                        description:
                            - A friendly name for the destination.
                            - This name should be unique across all destinations (regardless of type) within the data collection rule.
                        type: str
    data_flows:
        description:
            - Definition of which streams are sent to which destinations.
            - >-
              U(https://learn.microsoft.com/en-us/python/api/azure-mgmt-monitor/azure.mgmt.monitor.v2022_06_01.models.dataflow?view=azure-python)
        type: list
        elements: dict
        suboptions:
            streams:
                description:
                    - List of streams for this data flow.
                    - >-
                      U(https://learn.microsoft.com/en-us/rest/api/monitor/data-collection-rules/create?view=rest-monitor-2023-03-11&tabs=HTTP&tryIt=true#knowndataflowstreams)
                type: list
                elements: str
            destinations:
                description:
                    - List of destinations for this data flow.
                type: list
                elements: str
            transform_kql:
                description:
                    - The KQL query to transform stream data.
                type: str
            output_stream:
                description:
                    - The output stream of the transform. Only required if the transform changes data to a different stream.
                type: str
            built_in_transform:
                description:
                    - The builtIn transform to transform stream data.
                type: str
    state:
        description:
            - State of the data collection rule
            - Use C(present) for creating/updating a data collection rule.
            - Use C(absent) for deleting a data collection rule.
        default: present
        type: str
        choices:
            - present
            - absent
extends_documentation_fragment:
    - azure.azcollection.azure
    - azure.azcollection.azure_tags

author:
    - Klaas Demter (@Klaas-)
'''

EXAMPLES = '''
- name: Add a data collection rule
  azure.azcollection.azure_rm_monitordatacollectionrules:
    state: present
    name: data_collection_rule_name
    resource_group: resource_group_name
    location: westeurope
    kind: Linux
    description: This is an example description of a data collection rule
    data_sources:
      performance_counters:
        - name: perfCounterDataSource
          streams:
            - Microsoft-Perf
          sampling_frequency_in_seconds: 60
          counter_specifiers:
            - Processor(*)\\% Processor Time
            - Processor(*)\\% Idle Time
    destinations:
      log_analytics:
        - workspace_resource_id: "/subscriptions/<sub-id>/resourceGroups/rg/providers/Microsoft.OperationalInsights/workspaces/log_analytics_workspace_name"
          name: log_analytics_workspace_name
    data_flows:
      - destinations:
          - log_analytics_workspace_name
        streams:
          - Microsoft-Perf
    append_tags: false
    tags:
      ThisIsAnExampleTag: ExampleValue

- name: Add a data collection rule
  azure.azcollection.azure_rm_monitordatacollectionrules:
    state: present
    name: data_collection_rule_name
    resource_group: resource_group_name
    append_tags: true
    tags:
      ThisIsAnAddedExampleTag: ExampleValue

# Note this needs a DCR endpoint, not sure why, creating one via portal does not need that
# Also the table in your log analytics workspace has to already exist
- name: Add a data collection rule for collecting a custom log
  azure.azcollection.azure_rm_monitordatacollectionrules:
    name: data_collection_rule_name
    resource_group: resource_group_name
    location: westeurope
    kind: Linux
    data_sources:
      log_files:
        - file_patterns:
            - /var/log/dnf.rpm.log
          format: text
          name: Custom-Text-CustomLogs_CL
          streams:
            - Custom-Text-CustomLogs_CL
    destinations:
      log_analytics:
        - workspace_resource_id: "/subscriptions/<sub-id>/resourceGroups/rg/providers/Microsoft.OperationalInsights/workspaces/log_analytics_workspace_name"
          name: log_analytics_workspace_name
    data_flows:
      - destinations:
          - log_analytics_workspace_name
        output_stream: Custom-CustomLogs_CL
        streams:
          - Custom-Text-CustomLogs_CL
        transform_kql: source
    data_collection_endpoint_id: "/subscriptions/<sub-id>/resourceGroups/rg/providers/Microsoft.Insights/dataCollectionEndpoints/dcr-endpoint"
    stream_declarations:
      Custom-Text-CustomLogs_CL:
        columns:
          - name: TimeGenerated
            type: datetime
          - name: RawData
            type: string
          - name: FilePath
            type: string
          - name: Computer
            type: string


- name: Delete a data collection rule
  azure.azcollection.azure_rm_monitordatacollectionrules:
    state: absent
    name: data_collection_rule_name
    resource_group: resource_group_name
'''

RETURN = '''
datacollectionrule:
    description:
        - Details of the data collection rule
        - Is null on state==absent (data collection rule does not exist or will be deleted)
        - Assumes you make legal changes in check mode
    type: dict
    returned: always
    sample: {
        "data_flows": [...],
        "data_sources": {},
        "description": "Description of your data collection rule",
        "destinations": {},
        "etag": "str",
        "id": \
"/subscriptions/<sub-id>/resourceGroups/resource_group_name/providers/Microsoft.Insights/dataCollectionRules/data_collection_rule_name",
        "immutable_id": "dcr-xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
        "kind": "Linux",
        "location": "westeurope",
        "name": "data_collection_rule_name",
        "provisioning_state": "Succeeded",
        "system_data": {
            "created_at": "2025-01-01T00:00:00.000000Z",
            "created_by": "xxx@domain.tld",
            "created_by_type": "User",
            "last_modified_at": "2025-01-01T00:00:00.000000Z",
            "last_modified_by": "xxx@domain.tld",
            "last_modified_by_type": "User"
        },
        "tags": {},
        "type": "Microsoft.Insights/dataCollectionRules"
    }
'''

from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common_ext import AzureRMModuleBaseExt

try:
    from azure.core.exceptions import ResourceNotFoundError

except ImportError:
    # This is handled in azure_rm_common
    pass

AZURE_OBJECT_CLASS = 'DataCollectionRules'

data_flows_spec = dict(
    streams=dict(type='list', elements='str'),
    destinations=dict(type='list', elements='str'),
    transform_kql=dict(type='str'),
    output_stream=dict(type='str'),
    built_in_transform=dict(type='str'),
)

destinations_spec_log_analytics_options = dict(
    workspace_resource_id=dict(type='str'),
    name=dict(type='str')
)

destinations_spec_monitoring_accounts_options = dict(
    account_resource_id=dict(type='str'),
    name=dict(type='str')
)

destinations_spec_event_hubs_options = dict(
    event_hub_resource_id=dict(type='str'),
    name=dict(type='str')
)

destinations_spec_storage_blobs_options = dict(
    container_name=dict(type='str'),
    storage_account_resource_id=dict(type='str'),
    name=dict(type='str')
)

destinations_spec_storage_tables_options = dict(
    table_name=dict(type='str'),
    storage_account_resource_id=dict(type='str'),
    name=dict(type='str')
)

destinations_spec = dict(
    log_analytics=dict(type='list', elements='dict', options=destinations_spec_log_analytics_options),
    monitoring_accounts=dict(type='list', elements='dict', options=destinations_spec_monitoring_accounts_options),
    azure_monitor_metrics=dict(type='str'),
    event_hubs=dict(type='list', elements='dict', options=destinations_spec_event_hubs_options),
    event_hubs_direct=dict(type='list', elements='dict', options=destinations_spec_event_hubs_options),
    storage_blobs_direct=dict(type='list', elements='dict', options=destinations_spec_storage_blobs_options),
    storage_tables_direct=dict(type='list', elements='dict', options=destinations_spec_storage_tables_options),
    storage_accounts=dict(type='list', elements='dict', options=destinations_spec_storage_blobs_options)
)

data_sources_spec_performance_counters_options = dict(
    streams=dict(type='list', elements='str'),
    sampling_frequency_in_seconds=dict(type='int'),
    counter_specifiers=dict(type='list', elements='str'),
    name=dict(type='str')
)

data_sources_spec_windows_event_logs_options = dict(
    streams=dict(type='list', elements='str'),
    x_path_queries=dict(type='list', elements='str'),
    name=dict(type='str')
)

data_sources_spec_syslog_options = dict(
    streams=dict(type='list', elements='str'),
    facility_names=dict(type='list', elements='str'),
    log_levels=dict(type='list', elements='str'),
    name=dict(type='str')
)

data_sources_spec_extensions_options = dict(
    streams=dict(type='list', elements='str'),
    extension_name=dict(type='list', elements='str'),
    extension_settings=dict(type='str'),
    input_data_sources=dict(type='list', elements='str'),
    name=dict(type='str')
)

data_sources_spec_log_files_options = dict(
    streams=dict(type='list', elements='str'),
    file_patterns=dict(type='list', elements='str'),
    format=dict(type='str'),
    settings=dict(type='dict', options=dict(text=dict(type='str'))),
    name=dict(type='str')
)

data_sources_spec_iis_logs_options = dict(
    streams=dict(type='list', elements='str'),
    log_directories=dict(type='list', elements='str'),
    name=dict(type='str')
)

data_sources_spec_windows_firewall_logs_options = dict(
    streams=dict(type='list', elements='str'),
    name=dict(type='str')
)

data_sources_spec_prometheus_forwarder_options = dict(
    streams=dict(type='list', elements='str'),
    label_include_filter=dict(type='dict'),
    name=dict(type='str')
)

data_sources_spec_data_imports_event_hub_options = dict(
    name=dict(type='str'),
    consumer_group=dict(type='str'),
    stream=dict(type='str')
)

data_sources_spec_data_imports_options = dict(
    event_hub=dict(type='dict', options=data_sources_spec_data_imports_event_hub_options)
)

data_sources_spec = dict(
    performance_counters=dict(type='list', elements='dict', options=data_sources_spec_performance_counters_options),
    windows_event_logs=dict(type='list', elements='dict', options=data_sources_spec_windows_event_logs_options),
    syslog=dict(type='list', elements='dict', options=data_sources_spec_syslog_options),
    extensions=dict(type='list', elements='dict', options=data_sources_spec_extensions_options),
    log_files=dict(type='list', elements='dict', options=data_sources_spec_log_files_options),
    iis_logs=dict(type='list', elements='dict', options=data_sources_spec_iis_logs_options),
    windows_firewall_logs=dict(type='list', elements='dict', options=data_sources_spec_windows_firewall_logs_options),
    prometheus_forwarder=dict(type='list', elements='dict', options=data_sources_spec_prometheus_forwarder_options),
    # Platform telemetry options are currently same as windows_firewall_logs
    platform_telemetry=dict(type='list', elements='dict', options=data_sources_spec_windows_firewall_logs_options),
    data_imports=dict(type='dict', options=data_sources_spec_data_imports_options)
)


class AzureRMDataCollectionRules(AzureRMModuleBaseExt):
    """Information class for an Azure RM Data Collection Rules"""

    def __init__(self):
        self.module_arg_spec = dict(
            name=dict(type='str', required=True),
            resource_group=dict(type='str', required=True),
            location=dict(type='str'),
            kind=dict(type='str', choices=['Linux', 'Windows']),
            description=dict(type='str'),
            data_collection_endpoint_id=dict(type='str'),
            # https://github.com/ansible/ansible/issues/74001
            # Can't properly define this in arg spec
            stream_declarations=dict(type='dict'),
            data_sources=dict(type='dict', options=data_sources_spec),
            destinations=dict(type='dict', options=destinations_spec),
            data_flows=dict(type='list', elements='dict', options=data_flows_spec),
            state=dict(type='str', choices=['present', 'absent'], default='present')
        )

        self.name = None
        self.resource_group = None
        self.location = None
        self.kind = None
        self.description = None
        self.data_collection_endpoint_id = None
        self.stream_declarations = None
        self.data_sources = None
        self.destinations = None
        self.data_flows = None
        self.state = None
        self.tags = None
        self.log_path = None
        self.log_mode = None

        self.results = dict(
            changed=False,
            datacollectionrule=dict(),
            diff=dict(
                before=None,
                after=None
            )
        )

        super(AzureRMDataCollectionRules, self).__init__(derived_arg_spec=self.module_arg_spec,
                                                         supports_check_mode=True,
                                                         supports_tags=True)

    def exec_module(self, **kwargs):
        """Main module execution method"""

        for key in list(self.module_arg_spec.keys()) + ['tags']:
            if hasattr(self, key):
                setattr(self, key, kwargs[key])

        # Defaults for variables
        result = None
        result_compare = dict(compare=[])
        before_dict = None

        # Get current data collection rule if it exists
        before_dict = self.get_data_collection_rule()

        # Create dict from input, without None values
        # https://learn.microsoft.com/en-us/python/api/azure-mgmt-monitor/azure.mgmt.monitor.v2021_04_01.models.datacollectionruleresource?view=azure-python
        data_collection_rule_template = {
            "location": self.location,
            "kind": self.kind,
            "description": self.description,
            "data_collection_endpoint_id": self.data_collection_endpoint_id,
            "stream_declarations": self.stream_declarations,
            "data_sources": self.data_sources,
            "destinations": self.destinations,
            "data_flows": self.data_flows
        }
        # Filter out all None values
        data_collection_rule_input = {key: value for key, value in data_collection_rule_template.items() if value is not None}

        # Create/Update if state==present
        if self.state == 'present':
            if before_dict is None:
                # Data collection rule does not exist, create
                # On creation default to location of resource group unless otherwise noted in input variables
                if not self.location:
                    resource_group = self.get_resource_group(self.resource_group)
                    data_collection_rule_input['location'] = resource_group.location
                # On creation input == what we send to api
                data_collection_rule_update = data_collection_rule_input
                # Needs to be extended by tags if set
                if self.tags:
                    data_collection_rule_update['tags'] = self.tags
                self.results['changed'] = True
                if self.check_mode:
                    # Check mode, skipping actual creation
                    pass
                else:
                    create_response = self.create_or_update(data_collection_rule_update)
            else:
                # Data collection rule already exists, updating it
                # Dict for update is the union of existing object overwritten by input data
                data_collection_rule_update = before_dict | data_collection_rule_input

                # Enhanced with tags (special behaviour because of append_tags possibility)
                update_tags, update_tags_content = self.update_tags(before_dict.get('tags'))
                # Check if we need to update the data collection rule
                if update_tags or not self.default_compare({}, data_collection_rule_update, before_dict, '', result_compare):
                    data_collection_rule_update['tags'] = update_tags_content
                    # Need to create/update the Data collection rule; changed -> True
                    self.results['changed'] = True
                    if self.check_mode:
                        # Check mode, skipping actual creation
                        pass
                    else:
                        create_response = self.create_or_update(data_collection_rule_update)

            if self.check_mode or not self.results['changed']:
                # When object was not updated or when running in check mode
                # assume data_collection_rule_update is resulting object
                result = data_collection_rule_update
            else:
                # otherwise take resulting new object from response of create call
                result = create_response

        # Delete data collection rule if state is absent and it exists
        # if it doesn't exist, it's already absent
        elif self.state == 'absent' and before_dict is not None:
            self.results['changed'] = True
            if self.check_mode:
                # do not delete in check mode
                pass
            else:
                self.delete()

        self.results['diff']['before'] = before_dict
        self.results['diff']['after'] = result
        self.results['datacollectionrule'] = result

        return self.results

    def get_data_collection_rule(self):
        '''
        Gets the properties of the specified data collection rule.

        :return: List of Data Collection Rules
        '''
        self.log("Checking if data collection rule {0} in resource group {1} is present".format(self.name,
                                                                                                self.resource_group))

        result = None
        response = None

        try:
            response = self.monitor_management_client_data_collection_rules.data_collection_rules.get(data_collection_rule_name=self.name,
                                                                                                      resource_group_name=self.resource_group)
        except ResourceNotFoundError as ex:
            self.log("Could not find data collection rule {0} in resource group {1}".format(self.name, self.resource_group))
        if response:
            result = self.serialize_obj(response, AZURE_OBJECT_CLASS)

        return result

    def create_or_update(self, data_collection_rule_update):
        result = None
        response = None
        try:
            response = self.monitor_management_client_data_collection_rules.data_collection_rules.create(resource_group_name=self.resource_group,
                                                                                                         data_collection_rule_name=self.name,
                                                                                                         body=data_collection_rule_update,
                                                                                                         logging_enable=False)
        except Exception as ex:
            self.fail("Error creating or update data collection rule {0} in resource group {1}: {2}".format(self.name, self.resource_group, str(ex)))

        if response:
            result = self.serialize_obj(response, AZURE_OBJECT_CLASS)

        return result

    def delete(self):
        response = None
        try:
            response = self.monitor_management_client_data_collection_rules.data_collection_rules.delete(resource_group_name=self.resource_group,
                                                                                                         data_collection_rule_name=self.name)
        except Exception as ex:
            self.fail("Error deleting data collection rule {0} in resource group {1}: {2}".format(self.name, self.resource_group, str(ex)))

        return response


def main():
    """Main execution"""
    AzureRMDataCollectionRules()


if __name__ == '__main__':
    main()



#!/usr/bin/python
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, division, print_function


__metaclass__ = type


DOCUMENTATION = """
module: nxos_gir
extends_documentation_fragment:
- cisco.nxos.nxos
short_description: Trigger a graceful removal or insertion (GIR) of the switch.
description:
- Trigger a graceful removal or insertion (GIR) of the switch.
- GIR processing may take more than 2 minutes. Timeout settings are automatically
  extended to 200s when user timeout settings are insufficient.
version_added: 1.0.0
author:
- Gabriele Gerbino (@GGabriele)
notes:
- Tested against NXOSv 7.3.(0)D1(1) on VIRL
- Unsupported for Cisco MDS
- C(state) has effect only in combination with C(system_mode_maintenance_timeout)
  or C(system_mode_maintenance_on_reload_reset_reason).
- Using C(system_mode_maintenance) and C(system_mode_maintenance_dont_generate_profile)
  would make the module fail, but the system mode will be triggered anyway.
options:
  system_mode_maintenance:
    description:
    - When C(system_mode_maintenance=true) it puts all enabled protocols in maintenance
      mode (using the isolate command). When C(system_mode_maintenance=false) it puts
      all enabled protocols in normal mode (using the no isolate command).
    type: bool
  system_mode_maintenance_dont_generate_profile:
    description:
    - When C(system_mode_maintenance_dont_generate_profile=true) it prevents the dynamic
      searching of enabled protocols and executes commands configured in a maintenance-mode
      profile. Use this option if you want the system to use a maintenance-mode profile
      that you have created. When C(system_mode_maintenance_dont_generate_profile=false)
      it prevents the dynamic searching of enabled protocols and executes commands
      configured in a normal-mode profile. Use this option if you want the system
      to use a normal-mode profile that you have created.
    type: bool
  system_mode_maintenance_timeout:
    description:
    - Keeps the switch in maintenance mode for a specified number of minutes. Range
      is 5-65535.
    type: str
  system_mode_maintenance_shutdown:
    description:
    - Shuts down all protocols, vPC domains, and interfaces except the management
      interface (using the shutdown command). This option is disruptive while C(system_mode_maintenance)
      (which uses the isolate command) is not.
    type: bool
  system_mode_maintenance_on_reload_reset_reason:
    description:
    - Boots the switch into maintenance mode automatically in the event of a specified
      system crash. Note that not all reset reasons are applicable for all platforms.
      Also if reset reason is set to match_any, it is not idempotent as it turns on
      all reset reasons. If reset reason is match_any and state is absent, it turns
      off all the reset reasons.
    choices:
    - hw_error
    - svc_failure
    - kern_failure
    - wdog_timeout
    - fatal_error
    - lc_failure
    - match_any
    - manual_reload
    - any_other
    - maintenance
    type: str
  state:
    description:
    - Specify desired state of the resource.
    default: present
    choices:
    - present
    - absent
    type: str
"""

EXAMPLES = """
# Trigger system maintenance mode
- cisco.nxos.nxos_gir:
    system_mode_maintenance: true
    host: '{{ inventory_hostname }}'
    username: '{{ un }}'
    password: '{{ pwd }}'
# Trigger system normal mode
- cisco.nxos.nxos_gir:
    system_mode_maintenance: false
    host: '{{ inventory_hostname }}'
    username: '{{ un }}'
    password: '{{ pwd }}'
# Configure on-reload reset-reason for maintenance mode
- cisco.nxos.nxos_gir:
    system_mode_maintenance_on_reload_reset_reason: manual_reload
    state: present
    host: '{{ inventory_hostname }}'
    username: '{{ un }}'
    password: '{{ pwd }}'
# Add on-reload reset-reason for maintenance mode
- cisco.nxos.nxos_gir:
    system_mode_maintenance_on_reload_reset_reason: hw_error
    state: present
    host: '{{ inventory_hostname }}'
    username: '{{ un }}'
    password: '{{ pwd }}'
# Remove on-reload reset-reason for maintenance mode
- cisco.nxos.nxos_gir:
    system_mode_maintenance_on_reload_reset_reason: manual_reload
    state: absent
    host: '{{ inventory_hostname }}'
    username: '{{ un }}'
    password: '{{ pwd }}'
# Set timeout for maintenance mode
- cisco.nxos.nxos_gir:
    system_mode_maintenance_timeout: 30
    state: present
    host: '{{ inventory_hostname }}'
    username: '{{ un }}'
    password: '{{ pwd }}'
# Remove timeout for maintenance mode
- cisco.nxos.nxos_gir:
    system_mode_maintenance_timeout: 30
    state: absent
    host: '{{ inventory_hostname }}'
    username: '{{ un }}'
    password: '{{ pwd }}'
"""

RETURN = """
final_system_mode:
    description: describe the last system mode
    returned: verbose mode
    type: str
    sample: normal
updates:
    description: commands sent to the device
    returned: verbose mode
    type: list
    sample: ["terminal dont-ask", "system mode maintenance timeout 10"]
changed:
    description: check to see if a change was made on the device
    returned: always
    type: bool
    sample: true
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.nxos import (
    load_config,
    run_commands,
)


def get_system_mode(module):
    command = {"command": "show system mode", "output": "text"}
    body = run_commands(module, [command])[0]
    if body and "normal" in body.lower():
        mode = "normal"
    else:
        mode = "maintenance"
    return mode


def get_maintenance_timeout(module):
    command = {"command": "show maintenance timeout", "output": "text"}
    body = run_commands(module, [command])[0]
    timeout = body.split()[4]
    return timeout


def get_reset_reasons(module):
    command = {
        "command": "show maintenance on-reload reset-reasons",
        "output": "text",
    }
    body = run_commands(module, [command])[0]
    return body


def get_commands(module, state, mode):
    commands = list()
    if module.params["system_mode_maintenance"] is True and mode == "normal":
        commands.append("system mode maintenance")
    elif module.params["system_mode_maintenance"] is False and mode == "maintenance":
        commands.append("no system mode maintenance")

    elif (
        module.params["system_mode_maintenance_dont_generate_profile"] is True and mode == "normal"
    ):
        commands.append("system mode maintenance dont-generate-profile")
    elif (
        module.params["system_mode_maintenance_dont_generate_profile"] is False
        and mode == "maintenance"
    ):
        commands.append("no system mode maintenance dont-generate-profile")

    elif module.params["system_mode_maintenance_timeout"]:
        timeout = get_maintenance_timeout(module)
        if state == "present" and timeout != module.params["system_mode_maintenance_timeout"]:
            commands.append(
                "system mode maintenance timeout {0}".format(
                    module.params["system_mode_maintenance_timeout"],
                ),
            )
        elif state == "absent" and timeout == module.params["system_mode_maintenance_timeout"]:
            commands.append(
                "no system mode maintenance timeout {0}".format(
                    module.params["system_mode_maintenance_timeout"],
                ),
            )

    elif module.params["system_mode_maintenance_shutdown"] and mode == "normal":
        commands.append("system mode maintenance shutdown")
    elif module.params["system_mode_maintenance_shutdown"] is False and mode == "maintenance":
        commands.append("no system mode maintenance")

    elif module.params["system_mode_maintenance_on_reload_reset_reason"]:
        reset_reasons = get_reset_reasons(module)
        if (
            state == "present"
            and module.params["system_mode_maintenance_on_reload_reset_reason"].lower()
            not in reset_reasons.lower()
        ):
            commands.append(
                "system mode maintenance on-reload "
                "reset-reason {0}".format(
                    module.params["system_mode_maintenance_on_reload_reset_reason"],
                ),
            )
        elif (
            state == "absent"
            and module.params["system_mode_maintenance_on_reload_reset_reason"].lower()
            in reset_reasons.lower()
        ):
            commands.append(
                "no system mode maintenance on-reload "
                "reset-reason {0}".format(
                    module.params["system_mode_maintenance_on_reload_reset_reason"],
                ),
            )

    if commands:
        commands.insert(0, "terminal dont-ask")
    return commands


def main():
    argument_spec = dict(
        system_mode_maintenance=dict(required=False, type="bool"),
        system_mode_maintenance_dont_generate_profile=dict(required=False, type="bool"),
        system_mode_maintenance_timeout=dict(required=False, type="str"),
        system_mode_maintenance_shutdown=dict(required=False, type="bool"),
        system_mode_maintenance_on_reload_reset_reason=dict(
            required=False,
            choices=[
                "hw_error",
                "svc_failure",
                "kern_failure",
                "wdog_timeout",
                "fatal_error",
                "lc_failure",
                "match_any",
                "manual_reload",
                "any_other",
                "maintenance",
            ],
        ),
        state=dict(choices=["absent", "present"], default="present", required=False),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        mutually_exclusive=[
            [
                "system_mode_maintenance",
                "system_mode_maintenance_dont_generate_profile",
                "system_mode_maintenance_timeout",
                "system_mode_maintenance_shutdown",
                "system_mode_maintenance_on_reload_reset_reason",
            ],
        ],
        required_one_of=[
            [
                "system_mode_maintenance",
                "system_mode_maintenance_dont_generate_profile",
                "system_mode_maintenance_timeout",
                "system_mode_maintenance_shutdown",
                "system_mode_maintenance_on_reload_reset_reason",
            ],
        ],
        supports_check_mode=True,
    )

    warnings = list()

    state = module.params["state"]
    mode = get_system_mode(module)
    commands = get_commands(module, state, mode)
    changed = False
    if commands:
        if module.check_mode:
            module.exit_json(changed=True, commands=commands)
        else:
            load_config(module, commands)
            changed = True

    result = {}
    result["changed"] = changed
    if module._verbosity > 0:
        final_system_mode = get_system_mode(module)
        result["final_system_mode"] = final_system_mode
        result["updates"] = commands

    result["warnings"] = warnings

    module.exit_json(**result)


if __name__ == "__main__":
    main()



#
# -*- coding: utf-8 -*-
# Copyright 2025 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
#

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The nxos_l3_interfaces config file.
It is in this file where the current configuration (as dict)
is compared to the provided configuration (as dict) and the command set
necessary to bring the current configuration to its desired end-state is
created.
"""

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.resource_module import (
    ResourceModule,
)
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.utils import (
    dict_merge,
)

from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.facts.facts import (
    Facts,
)
from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.rm_templates.l3_interfaces import (
    L3_interfacesTemplate,
)


class L3_interfaces(ResourceModule):
    """
    The nxos_l3_interfaces config class
    """

    def __init__(self, module):
        super(L3_interfaces, self).__init__(
            empty_fact_val={},
            facts_module=Facts(module),
            module=module,
            resource="l3_interfaces",
            tmplt=L3_interfacesTemplate(),
        )
        self.parsers = [
            "mac_address",
            "bandwidth",
            "dot1q",
            "evpn_multisite_tracking",
            "redirects",
            "unreachables",
            "proxy_arp",
            "port_unreachable",
            "verify",
            "dhcp.ipv4",
            "dhcp.ipv4.option82",
            "dhcp.ipv4.information",
            "dhcp.ipv4.subnet_selection",
            "dhcp.ipv4.source_interface",
            "ipv6_redirects",
            "ipv6_unreachables",
            "ipv6_verify",
            "dhcp.ipv6",
            "dhcp.ipv6.source_interface",
        ]

    def execute_module(self):
        """Execute the module

        :rtype: A dictionary
        :returns: The result from module execution
        """
        if self.state not in ["parsed", "gathered"]:
            self.generate_commands()
            self.run_commands()
        return self.result

    def generate_commands(self):
        """Generate configuration commands to send based on
        want, have and desired state.
        """

        wantd = self.convert_list_to_dict(self.want)
        haved = self.convert_list_to_dict(self.have)

        # if state is merged, merge want onto have and then compare
        if self.state == "merged":
            wantd = dict_merge(haved, wantd)

        # if state is deleted, empty out wantd and set haved to wantd
        if self.state == "deleted":
            haved = {k: v for k, v in haved.items() if k in wantd or not wantd}
            wantd = {}

        # remove superfluous config for overridden and deleted
        if self.state in ["overridden", "deleted"]:
            for k, have in haved.items():
                if k not in wantd:
                    self._compare(want={}, have=have)

        for k, want in wantd.items():
            self._compare(want=want, have=haved.pop(k, {}))

    def _compare(self, want, have):
        """Leverages the base class `compare()` method and
        populates the list of commands to be run by comparing
        the `want` and `have` data with the `parsers` defined
        for the L3_interfaces network resource.
        """
        begin = len(self.commands)
        want_without_name = want.copy()
        want_without_name.pop("name", None)
        pre_pop_want = bool(want_without_name)
        want_redirects = want.pop("redirects", None)
        have_redirects = have.pop("redirects", None)
        self.handle_redirects(want_redirects, have_redirects, "redirects", pre_pop_want)
        want_redirects = want.pop("ipv6_redirects", None)
        have_redirects = have.pop("ipv6_redirects", None)
        self.handle_redirects(want_redirects, have_redirects, "ipv6_redirects", pre_pop_want)
        self.compare(parsers=self.parsers, want=want, have=have)
        self._compare_complex_attrs(want, have)
        if len(self.commands) != begin:
            self.commands.insert(begin, self._tmplt.render(want or have, "name", False))

    def convert_list_to_dict(self, data):
        def list_to_dict(dict_list, outer_key_priority):
            output = {}
            for entry in dict_list:
                if not isinstance(entry, dict) or not entry:
                    continue

                outer_key = None

                for k in outer_key_priority:
                    if k in entry:
                        val = entry[k]
                        outer_key = k if isinstance(val, bool) else str(val)
                        break

                if outer_key is not None:
                    output[outer_key] = entry

            return output

        result = {}

        iterable = ((entry["name"], entry) for entry in data if "name" in entry)

        for iface_name, iface_data in iterable:
            iface_result = iface_data.copy()

            ipv4_value_addresses = iface_result.get("ipv4", [])
            ipv6_value_addresses = iface_result.get("ipv6", [])

            ipv4_value_dhcp_relay_address = (
                iface_result.get("dhcp", {}).get("ipv4", {}).get("relay", {}).get("address", [])
            )
            ipv6_value_dhcp_relay_address = (
                iface_result.get("dhcp", {}).get("ipv6", {}).get("relay", {}).get("address", [])
            )

            if ipv4_value_addresses:
                iface_result["ipv4"] = list_to_dict(
                    ipv4_value_addresses,
                    ["address", "ip_network_mask"],
                )
            if ipv4_value_dhcp_relay_address:
                iface_result["dhcp"]["ipv4"]["relay"]["address"] = list_to_dict(
                    ipv4_value_dhcp_relay_address,
                    ["relay_ip"],
                )
            if ipv6_value_addresses:
                iface_result["ipv6"] = list_to_dict(
                    ipv6_value_addresses,
                    [
                        "address",
                    ],
                )
            if ipv6_value_dhcp_relay_address:
                iface_result["dhcp"]["ipv6"]["relay"]["address"] = list_to_dict(
                    ipv6_value_dhcp_relay_address,
                    ["relay_ip"],
                )

            result[iface_name] = iface_result

        return result

    def _compare_complex_attrs(self, want, have):
        """Compare complex attributes"""
        want_ipv4_value_address = want.get("ipv4", {})
        have_ipv4_value_address = have.get("ipv4", {})
        want_ipv6_value_address = want.get("ipv6", {})
        have_ipv6_value_address = have.get("ipv6", {})

        self.compare_lists(want_ipv4_value_address, have_ipv4_value_address, "ipv4.address")
        self.compare_lists(want_ipv6_value_address, have_ipv6_value_address, "ipv6.address")

        want_ipv4_value_relay_address = (
            want.get("dhcp", {}).get("ipv4", {}).get("relay", {}).get("address", {})
        )
        have_ipv4_value_relay_address = (
            have.get("dhcp", {}).get("ipv4", {}).get("relay", {}).get("address", {})
        )
        self.compare_lists(
            want_ipv4_value_relay_address,
            have_ipv4_value_relay_address,
            "dhcp.ipv4.address",
        )

        want_ipv6_value_relay_address = (
            want.get("dhcp", {}).get("ipv6", {}).get("relay", {}).get("address", {})
        )
        have_ipv6_value_relay_address = (
            have.get("dhcp", {}).get("ipv6", {}).get("relay", {}).get("address", {})
        )
        self.compare_lists(
            want_ipv6_value_relay_address,
            have_ipv6_value_relay_address,
            "dhcp.ipv6.address",
        )

    def compare_lists(self, wanted, haved, parser):
        """Compare list items in ipv4 and ipv6"""
        ip_key = parser.split(".")[0]
        parser_key = parser.split(".")[1]
        for key, want_value in wanted.items():
            have_value = haved.pop(key, {})
            self.compare(
                parsers=[parser],
                want={ip_key: {parser_key: want_value}},
                have={ip_key: {parser_key: have_value}},
            )
        for key, have_value in haved.items():
            self.compare(parsers=[parser], want={}, have={ip_key: {parser_key: have_value}})

    def handle_redirects(self, want_redirects, have_redirects, parser, want):
        if want_redirects is None and have_redirects is None:
            if self.state == "replaced" or (self.state == "overridden" and want):
                self.addcmd({parser: True}, parser, True)
        else:
            if want_redirects is True and have_redirects is False:
                self.addcmd({parser: want_redirects}, parser, not want_redirects)
            elif want_redirects is False and have_redirects is None:
                self.addcmd({parser: not want_redirects}, parser, not want_redirects)
            elif want_redirects is None and have_redirects is False:
                if self.state in ["overridden", "deleted"] and not want:
                    self.addcmd({parser: not have_redirects}, parser, have_redirects)



#!/usr/bin/python
# Copyright (c) 2019, Patrick Pichler <ppichler+ansible@mgit.at>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: openssl_signature_info
version_added: 1.1.0
short_description: Verify signatures with openssl
description:
  - This module allows one to verify a signature for a file by a certificate.
  - The module uses the cryptography Python library.
author:
  - Patrick Pichler (@aveexy)
  - Markus Teufelberger (@MarkusTeufelberger)
extends_documentation_fragment:
  - community.crypto._attributes
  - community.crypto._attributes.info_module
  - community.crypto._attributes.idempotent_not_modify_state
  - community.crypto._cryptography_dep.minimum
options:
  path:
    description:
      - The signed file to verify.
      - This file will only be read and not modified.
    type: path
    required: true
  certificate_path:
    description:
      - The path to the certificate used to verify the signature.
      - Either O(certificate_path) or O(certificate_content) must be specified, but not both.
    type: path
  certificate_content:
    description:
      - The content of the certificate used to verify the signature.
      - Either O(certificate_path) or O(certificate_content) must be specified, but not both.
    type: str
  signature:
    description: Base64 encoded signature.
    type: str
    required: true
  select_crypto_backend:
    description:
      - Determines which crypto backend to use.
      - The default choice is V(auto), which tries to use C(cryptography) if available.
      - If set to V(cryptography), will try to use the L(cryptography,https://cryptography.io/) library.
      - Note that with community.crypto 3.0.0, all values behave the same.
        This option will be deprecated in a later version.
        We recommend to not set it explicitly.
    type: str
    default: auto
    choices: [auto, cryptography]
seealso:
  - module: community.crypto.openssl_signature
  - module: community.crypto.x509_certificate
"""

EXAMPLES = r"""
---
- name: Sign example file
  community.crypto.openssl_signature:
    privatekey_path: private.key
    path: /tmp/example_file
  register: sig

- name: Verify signature of example file
  community.crypto.openssl_signature_info:
    certificate_path: cert.pem
    path: /tmp/example_file
    signature: "{{ sig.signature }}"
  register: verify

- name: Make sure the signature is valid
  ansible.builtin.assert:
    that:
      - verify.valid
"""

RETURN = r"""
valid:
  description: V(true) means the signature was valid for the given file, V(false) means it was not.
  returned: success
  type: bool
"""

import base64
import os
import typing as t

from ansible_collections.community.crypto.plugins.module_utils._cryptography_dep import (
    COLLECTION_MINIMUM_CRYPTOGRAPHY_VERSION,
    assert_required_cryptography_version,
)
from ansible_collections.community.crypto.plugins.module_utils._version import (
    LooseVersion,
)

MINIMAL_CRYPTOGRAPHY_VERSION = COLLECTION_MINIMUM_CRYPTOGRAPHY_VERSION

try:
    import cryptography
    import cryptography.hazmat.primitives.asymmetric.padding
    import cryptography.hazmat.primitives.hashes

except ImportError:
    CRYPTOGRAPHY_VERSION = LooseVersion("0.0")  # pylint: disable=invalid-name
else:
    # pylint: disable-next=invalid-name
    CRYPTOGRAPHY_VERSION = LooseVersion(cryptography.__version__)

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.crypto.plugins.module_utils._crypto.basic import (
    OpenSSLObjectError,
)
from ansible_collections.community.crypto.plugins.module_utils._crypto.support import (
    OpenSSLObject,
    load_certificate,
)


class SignatureInfoBase(OpenSSLObject):
    def __init__(self, module: AnsibleModule) -> None:
        super().__init__(
            path=module.params["path"],
            state="present",
            force=False,
            check_mode=module.check_mode,
        )

        self.module = module
        self.signature: str = module.params["signature"]
        self.certificate_path: str | None = module.params["certificate_path"]
        certificate_content: str | None = module.params["certificate_content"]
        if certificate_content is not None:
            self.certificate_content: bytes | None = certificate_content.encode("utf-8")
        else:
            self.certificate_content = None

    def generate(self, module: AnsibleModule) -> None:
        # Empty method because OpenSSLObject wants this
        pass

    def dump(self) -> dict[str, t.Any]:
        # Empty method because OpenSSLObject wants this
        return {}


# Implementation with using cryptography
class SignatureInfoCryptography(SignatureInfoBase):
    def __init__(self, module: AnsibleModule) -> None:
        super().__init__(module)

    def run(self) -> dict[str, t.Any]:
        _padding = cryptography.hazmat.primitives.asymmetric.padding.PKCS1v15()
        _hash = cryptography.hazmat.primitives.hashes.SHA256()

        result: dict[str, t.Any] = {}

        try:
            with open(self.path, "rb") as f:
                _in = f.read()

            _signature = base64.b64decode(self.signature)
            certificate = load_certificate(
                path=self.certificate_path,
                content=self.certificate_content,
            )
            public_key = certificate.public_key()
            verified = False
            valid = False

            try:
                if isinstance(
                    public_key,
                    cryptography.hazmat.primitives.asymmetric.dsa.DSAPublicKey,
                ):
                    public_key.verify(_signature, _in, _hash)
                    verified = True
                    valid = True

                elif isinstance(
                    public_key,
                    cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePublicKey,
                ):
                    public_key.verify(
                        _signature,
                        _in,
                        cryptography.hazmat.primitives.asymmetric.ec.ECDSA(_hash),
                    )
                    verified = True
                    valid = True

                elif isinstance(
                    public_key,
                    (
                        cryptography.hazmat.primitives.asymmetric.ed25519.Ed25519PublicKey,
                        cryptography.hazmat.primitives.asymmetric.ed448.Ed448PublicKey,
                    ),
                ):
                    public_key.verify(_signature, _in)
                    verified = True
                    valid = True

                elif isinstance(
                    public_key,
                    cryptography.hazmat.primitives.asymmetric.rsa.RSAPublicKey,
                ):
                    public_key.verify(_signature, _in, _padding, _hash)
                    verified = True
                    valid = True
            except cryptography.exceptions.InvalidSignature:
                verified = True
                valid = False

            if not verified:
                self.module.fail_json(
                    msg=f"Unsupported key type. Your cryptography version is {CRYPTOGRAPHY_VERSION}"
                )
            result["valid"] = valid
            return result

        except Exception as e:
            raise OpenSSLObjectError(e) from e


def main() -> t.NoReturn:
    module = AnsibleModule(
        argument_spec={
            "certificate_path": {"type": "path"},
            "certificate_content": {"type": "str"},
            "path": {"type": "path", "required": True},
            "signature": {"type": "str", "required": True},
            "select_crypto_backend": {
                "type": "str",
                "choices": ["auto", "cryptography"],
                "default": "auto",
            },
        },
        mutually_exclusive=(["certificate_path", "certificate_content"],),
        required_one_of=(["certificate_path", "certificate_content"],),
        supports_check_mode=True,
    )

    if not os.path.isfile(module.params["path"]):
        module.fail_json(
            name=module.params["path"],
            msg=f"The file {module.params['path']} does not exist",
        )

    assert_required_cryptography_version(
        module, minimum_cryptography_version=MINIMAL_CRYPTOGRAPHY_VERSION
    )

    try:
        _sign = SignatureInfoCryptography(module)

        result = _sign.run()

        module.exit_json(**result)
    except OpenSSLObjectError as exc:
        module.fail_json(msg=str(exc))


if __name__ == "__main__":
    main()



# Copyright (c) 2019, Evgeniy Krysanov <evgeniy.krysanov@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this doc fragment is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations


class ModuleDocFragment:
    # Standard documentation fragment
    DOCUMENTATION = r"""
options:
  client_id:
    description:
      - The OAuth consumer key.
      - If not set the environment variable E(BITBUCKET_CLIENT_ID) is used.
    type: str
  client_secret:
    description:
      - The OAuth consumer secret.
      - If not set the environment variable E(BITBUCKET_CLIENT_SECRET) is used.
    type: str
  user:
    description:
      - The username.
      - If not set the environment variable E(BITBUCKET_USERNAME) is used.
      - O(ignore:username) is an alias of O(user) since community.general 6.0.0. It was an alias of O(workspace) before.
    type: str
    version_added: 4.0.0
    aliases: [username]
  password:
    description:
      - The App password.
      - If not set the environment variable E(BITBUCKET_PASSWORD) is used.
    type: str
    version_added: 4.0.0
notes:
  - Bitbucket OAuth consumer key and secret can be obtained from Bitbucket profile -> Settings -> Access Management -> OAuth.
  - Bitbucket App password can be created from Bitbucket profile -> Personal Settings -> App passwords.
  - If both OAuth and Basic Auth credentials are passed, OAuth credentials take precedence.
"""



# Copyright (c) 2020 Cisco and/or its affiliates.
# Copyright: (c) 2020, Shreyas Srish (@shrsr) <ssrish@cisco.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import absolute_import, division, print_function

__metaclass__ = type

DOCUMENTATION = """
---
name: aci
author:
- Shreyas Srish (@shrsr)
short_description: Ansible ACI HTTPAPI Plugin.
description:
  - This ACI plugin provides the HTTPAPI methods needed to initiate
    a connection to the APIC, send API requests and process the
    response from the controller.
"""

import ast
import base64
import json
import os
import re

from ansible.module_utils._text import to_text, to_native
from ansible.module_utils.connection import ConnectionError
from ansible.plugins.httpapi import HttpApiBase
from copy import copy, deepcopy

# Optional, only used for APIC signature-based authentication
try:
    from OpenSSL.crypto import FILETYPE_PEM, load_privatekey, sign

    HAS_OPENSSL = True
except ImportError:
    HAS_OPENSSL = False

# Signature-based authentication using cryptography
try:
    from cryptography.hazmat.primitives import serialization, hashes
    from cryptography.hazmat.primitives.asymmetric import padding
    from cryptography.hazmat.backends import default_backend

    HAS_CRYPTOGRAPHY = True
except ImportError:
    HAS_CRYPTOGRAPHY = False

CONNECTION_MAP = {"username": "remote_user", "timeout": "persistent_command_timeout"}
RESET_KEYS = ["username", "password", "port"]
CONNECTION_KEYS = RESET_KEYS + ["timeout", "use_proxy", "use_ssl", "validate_certs"]


class HttpApi(HttpApiBase):
    def __init__(self, *args, **kwargs):
        super(HttpApi, self).__init__(*args, **kwargs)
        self.params = None
        self.result = {}
        self.backup_hosts = None
        self.connection_error_check = False
        self.connection_parameters = {}
        self.current_host = None
        self.provided_hosts = None
        self.inventory_hosts = None

    def set_params(self, params):
        self.params = params

    # Login function is executed until connection to a host is established or until all the hosts in the list are exhausted
    def login(self, username, password):
        """Log in to APIC"""
        # Perform login request
        self.connection.queue_message("debug", "Establishing login for {0} to {1}".format(username, self.connection.get_option("host")))
        method = "POST"
        path = "/api/aaaLogin.json"
        payload = {"aaaUser": {"attributes": {"name": username, "pwd": password}}}
        data = json.dumps(payload)
        self.connection._connected = True
        try:
            response, response_data = self.connection.send(path, data, method=method)
            self.connection._auth = {"Cookie": response.headers.get("Set-Cookie")}
            self.connection.queue_message("debug", "Connection to {0} was successful".format(self.connection.get_option("host")))
        except Exception as exc_login:
            self.connection._connected = False
            exc_login.path = path
            raise

    def set_parameters(self):
        connection_parameters = {}
        for key in CONNECTION_KEYS:
            value = self.params.get(key) if self.params.get(key) is not None else self.connection.get_option(CONNECTION_MAP.get(key, key))
            if key == "username" and value is None:
                value = "admin"
            self.connection.set_option(CONNECTION_MAP.get(key, key), value)
            if key == "timeout" and self.connection.get_option("persistent_connect_timeout") <= value:
                self.connection.set_option("persistent_connect_timeout", value + 30)

            connection_parameters[key] = value
            if self.connection_parameters and value != self.connection_parameters.get(key) and key in RESET_KEYS:
                self.connection._connected = False
                self.connection.queue_message("debug", "Re-setting connection due to change in the {0}".format(key))

        if self.params.get("private_key") is not None:
            self.connection.set_option("session_key", None)
            connection_parameters["certificate_name"] = self.params.get("certificate_name")
            connection_parameters["private_key"] = self.params.get("private_key")
        elif self.connection.get_option("session_key") is not None and self.params.get("password") is None:
            connection_parameters["certificate_name"] = list(self.connection.get_option("session_key").keys())[0]
            connection_parameters["private_key"] = list(self.connection.get_option("session_key").values())[0]
        else:
            if self.connection_parameters.get("private_key") is not None:
                self.connection._connected = False
                self.connection.queue_message(
                    "debug", "Re-setting connection due to change from private/session key authentication to password authentication"
                )
            self.connection.set_option("session_key", None)
            connection_parameters["private_key"] = None
            connection_parameters["certificate_name"] = None

        if self.connection_parameters != connection_parameters:
            self.connection_parameters = copy(connection_parameters)

        self.set_hosts()

    def set_hosts(self):
        if self.params.get("host") is not None:
            hosts = ast.literal_eval(self.params.get("host")) if "[" in self.params.get("host") else self.params.get("host").split(",")
        else:
            if self.inventory_hosts is None:
                self.inventory_hosts = re.sub(r"[[\]]", "", self.connection.get_option("host")).split(",")
            hosts = self.inventory_hosts

        if self.provided_hosts is None:
            self.provided_hosts = deepcopy(hosts)
            self.connection.queue_message("debug", "Provided Hosts: {0}".format(self.provided_hosts))
            self.backup_hosts = deepcopy(hosts)
            self.current_host = self.backup_hosts.pop(0)
            self.connection.queue_message("debug", "Initializing operation on {0}".format(self.current_host))
        elif self.provided_hosts != hosts:
            self.provided_hosts = deepcopy(hosts)
            self.connection.queue_message("debug", "Provided Hosts have changed: {0}".format(self.provided_hosts))
            self.backup_hosts = deepcopy(hosts)
            try:
                self.backup_hosts.pop(self.backup_hosts.index(self.current_host))
                self.connection.queue_message("debug", "Connected host {0} found in the provided hosts. Continuing with it.".format(self.current_host))
            except Exception:
                self.current_host = self.backup_hosts.pop(0)
                self.connection._connected = False
                self.connection.queue_message("debug", "Initializing operation on {0}".format(self.current_host))
        self.connection.set_option("host", self.current_host)

    # One API call is made via each call to send_request from aci.py in module_utils
    # As long as a host is active in the list the API call will go through
    def send_request(self, method, path, data):
        """This method handles all APIC REST API requests other than login"""

        self.set_parameters()

        if self.connection_parameters.get("private_key") is not None:
            try:
                self.connection._auth = {"Cookie": "{0}".format(self.cert_auth(method, path, data).get("Cookie"))}
                self.connection._connected = True
            except Exception as exc_response:
                self.connection._connected = False
                return self._return_info("", method, self.validate_url(self.connection._url + path), str(exc_response))

        try:
            if self.connection._connected is False:
                self.login(self.connection.get_option("remote_user"), self.connection.get_option("password"))
            self.connection.queue_message("debug", "Sending {0} request to {1}".format(method, self.connection._url + path))
            response, response_data = self.connection.send(path, data, method=method)
            self.connection.queue_message(
                "debug", "Received response from {0} for {1} operation with HTTP: {2}".format(self.connection.get_option("host"), method, response.getcode())
            )
        except Exception as exc_response:
            self.connection.queue_message("debug", "Connection to {0} has failed: {1}".format(self.connection.get_option("host"), exc_response))
            if len(self.backup_hosts) == 0:
                self.provided_hosts = None
                self.connection._connected = False
                error = dict(
                    code=-1, text="No hosts left in the cluster to continue operation! Error on final host {0}".format(self.connection.get_option("host"))
                )
                if "path" in dir(exc_response):
                    path = exc_response.path
                return self._return_info("", method, self.validate_url(self.connection._url + path), str(exc_response), error=error)
            else:
                self.current_host = self.backup_hosts.pop(0)
                self.connection.queue_message("debug", "Switching host from {0} to {1}".format(self.connection.get_option("host"), self.current_host))
                self.connection.set_option("host", self.current_host)
            # recurse through function for retrying the request
            return self.send_request(method, path, data)
        # return statement executed upon each successful response from the request function
        return self._verify_response(response, method, path, response_data)

    # Built-in-function
    def handle_httperror(self, exc):
        self.connection.queue_message("debug", "Failed to receive response from {0} with {1}".format(self.connection.get_option("host"), exc))
        if exc.code == 401:
            raise ConnectionError(exc)
        elif exc.code == 403 and self.connection_parameters.get("private_key") is None:
            self.connection._auth = None
            self.login(self.connection.get_option("remote_user"), self.connection.get_option("password"))
            return True
        return exc

    def validate_url(self, url):
        validated_url = re.match(r"^.*?\.json|^.*?\.xml", url).group(0)
        if self.connection_parameters.get("port") is None:
            return validated_url.replace(re.match(r"(https?:\/\/.*)(:\d*)\/?(.*)", url).group(2), "")
        else:
            return validated_url

    def _verify_response(self, response, method, path, response_data):
        """Process the return code and response object from APIC"""
        response_value = self._get_response_value(response_data)
        response_code = response.getcode()
        path = self.validate_url(response.url)
        # Response check to remain consistent with fetch_url's response
        if str(response) == "HTTP Error 400: Bad Request":
            msg = "{0}".format(response)
        else:
            msg = "{0} ({1} bytes)".format(response.msg, len(response_value))
        return self._return_info(response_code, method, path, msg, respond_data=response_value)

    def _get_response_value(self, response_data):
        """Extract string data from response_data returned from APIC"""
        return to_text(response_data.getvalue())

    def _response_to_json(self, response_text):
        """Convert response_text to json format"""
        try:
            return json.loads(response_text) if response_text else {}
        # JSONDecodeError only available on Python 3.5+
        except Exception:
            return "Invalid JSON response: {0}".format(response_text)

    def _return_info(self, response_code, method, path, msg, respond_data=None, error=None):
        """Format success/error data and return with consistent format"""
        info = {}
        info["status"] = response_code
        info["method"] = method
        info["url"] = path
        info["msg"] = msg
        if error is not None:
            info["error"] = error
        else:
            info["error"] = {}
        # Response check to trigger key error if response_data is invalid
        if respond_data is not None:
            info["body"] = respond_data
        return info

    def cert_auth(self, method, path, payload=""):
        """Perform APIC signature-based authentication, not the expected SSL client certificate authentication."""

        if payload is None:
            payload = ""

        headers = dict()

        try:
            if HAS_CRYPTOGRAPHY:
                key = self.connection_parameters.get("private_key").encode()
                sig_key = serialization.load_pem_private_key(
                    key,
                    password=None,
                    backend=default_backend(),
                )
            else:
                sig_key = load_privatekey(FILETYPE_PEM, self.connection_parameters.get("private_key"))
        except Exception:
            private_key_file_path = os.path.abspath(os.path.join(self.params.get("working_directory"), self.connection_parameters.get("private_key")))
            if os.path.exists(private_key_file_path):
                try:
                    permission = "r"
                    if HAS_CRYPTOGRAPHY:
                        permission = "rb"
                    with open(private_key_file_path, permission) as fh:
                        private_key_content = fh.read()
                except Exception:
                    raise ConnectionError("Cannot open private key file {0}".format(private_key_file_path))
                try:
                    if HAS_CRYPTOGRAPHY:
                        sig_key = serialization.load_pem_private_key(private_key_content, password=None, backend=default_backend())
                    else:
                        sig_key = load_privatekey(FILETYPE_PEM, private_key_content)
                except Exception:
                    raise ConnectionError("Cannot load private key file {0}".format(self.connection_parameters.get("private_key")))
                if self.connection_parameters.get("certificate_name") is None:
                    self.connection_parameters["certificate_name"] = os.path.basename(os.path.splitext(self.connection_parameters.get("private_key"))[0])
            else:
                raise ConnectionError(
                    "Provided private key {0} does not appear to be a private key or provided file does not exist.".format(private_key_file_path)
                )
        if self.connection_parameters.get("certificate_name") is None:
            self.connection_parameters["certificate_name"] = self.connection.get_option("remote_user")
        sig_request = method + path + payload
        if HAS_CRYPTOGRAPHY:
            sig_signature = sig_key.sign(sig_request.encode(), padding.PKCS1v15(), hashes.SHA256())
        else:
            sig_signature = sign(sig_key, sig_request, "sha256")
        sig_dn = "uni/userext/user-{0}/usercert-{1}".format(self.connection.get_option("remote_user"), self.connection_parameters.get("certificate_name"))
        headers["Cookie"] = (
            "APIC-Certificate-Algorithm=v1.0; "
            + "APIC-Certificate-DN={0}; ".format(sig_dn)
            + "APIC-Certificate-Fingerprint=fingerprint; "
            + "APIC-Request-Signature={0}".format(to_native(base64.b64encode(sig_signature)))
        )
        return headers



# Copyright (c) 2020, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import typing as t

from ansible_collections.community.crypto.plugins.module_utils._crypto.cryptography_crl import (
    TIMESTAMP_FORMAT,
    cryptography_decode_revoked_certificate,
    cryptography_dump_revoked,
    cryptography_get_signature_algorithm_oid_from_crl,
)
from ansible_collections.community.crypto.plugins.module_utils._crypto.cryptography_support import (
    cryptography_oid_to_name,
)
from ansible_collections.community.crypto.plugins.module_utils._crypto.pem import (
    identify_pem_format,
)
from ansible_collections.community.crypto.plugins.module_utils._cryptography_dep import (
    COLLECTION_MINIMUM_CRYPTOGRAPHY_VERSION,
    assert_required_cryptography_version,
)

if t.TYPE_CHECKING:
    from ansible.module_utils.basic import AnsibleModule  # pragma: no cover

    from ansible_collections.community.crypto.plugins.plugin_utils._action_module import (  # pragma: no cover
        AnsibleActionModule,
    )
    from ansible_collections.community.crypto.plugins.plugin_utils._filter_module import (  # pragma: no cover
        FilterModuleMock,
    )

    GeneralAnsibleModule = t.Union[  # noqa: UP007
        AnsibleModule, AnsibleActionModule, FilterModuleMock
    ]  # pragma: no cover


# crypto_utils

MINIMAL_CRYPTOGRAPHY_VERSION = COLLECTION_MINIMUM_CRYPTOGRAPHY_VERSION

try:
    from cryptography import x509
except ImportError:
    pass


class CRLInfoRetrieval:
    def __init__(
        self,
        *,
        module: GeneralAnsibleModule,
        content: bytes,
        list_revoked_certificates: bool = True,
    ) -> None:
        # content must be a bytes string
        self.module = module
        self.content = content
        self.list_revoked_certificates = list_revoked_certificates
        self.name_encoding = module.params.get("name_encoding", "ignore")

    def get_info(self) -> dict[str, t.Any]:
        crl_pem = identify_pem_format(self.content)
        try:
            if crl_pem:
                crl = x509.load_pem_x509_crl(self.content)
            else:
                crl = x509.load_der_x509_crl(self.content)
        except ValueError as e:
            self.module.fail_json(msg=f"Error while decoding CRL: {e}")

        result: dict[str, t.Any] = {
            "changed": False,
            "format": "pem" if crl_pem else "der",
            "last_update": None,
            "next_update": None,
            "digest": None,
            "issuer_ordered": None,
            "issuer": None,
        }

        result["last_update"] = crl.last_update.strftime(TIMESTAMP_FORMAT)
        result["next_update"] = (
            crl.next_update.strftime(TIMESTAMP_FORMAT) if crl.next_update else None
        )
        result["digest"] = cryptography_oid_to_name(
            cryptography_get_signature_algorithm_oid_from_crl(crl)
        )
        issuer = []
        for attribute in crl.issuer:
            issuer.append([cryptography_oid_to_name(attribute.oid), attribute.value])
        result["issuer_ordered"] = issuer
        issuer_dict = {}
        for k, v in issuer:
            issuer_dict[k] = v
        result["issuer"] = issuer_dict
        if self.list_revoked_certificates:
            result["revoked_certificates"] = []
            for cert in crl:
                entry = cryptography_decode_revoked_certificate(cert)
                result["revoked_certificates"].append(
                    cryptography_dump_revoked(entry, idn_rewrite=self.name_encoding)
                )

        return result


def get_crl_info(
    *,
    module: GeneralAnsibleModule,
    content: bytes,
    list_revoked_certificates: bool = True,
) -> dict[str, t.Any]:
    assert_required_cryptography_version(
        module, minimum_cryptography_version=MINIMAL_CRYPTOGRAPHY_VERSION
    )
    info = CRLInfoRetrieval(
        module=module,
        content=content,
        list_revoked_certificates=list_revoked_certificates,
    )
    return info.get_info()


__all__ = ("CRLInfoRetrieval", "get_crl_info")



# This code is part of the Ansible collection community.docker, but is an independent component.
# This particular file, and this file only, is based on the Docker SDK for Python (https://github.com/docker/docker-py/)
#
# Copyright (c) 2016-2022 Docker, Inc.
#
# It is licensed under the Apache 2.0 license (see LICENSES/Apache-2.0.txt in this collection)
# SPDX-License-Identifier: Apache-2.0

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

from .._import_helper import HTTPAdapter as _HTTPAdapter


class BaseHTTPAdapter(_HTTPAdapter):
    def close(self) -> None:
        # pylint finds our HTTPAdapter stub instead of requests.adapters.HTTPAdapter:
        # pylint: disable-next=no-member
        super().close()
        if hasattr(self, "pools"):
            self.pools.clear()

    # Hotfix for requests 2.32.0 and 2.32.1: its commit
    # https://github.com/psf/requests/commit/c0813a2d910ea6b4f8438b91d315b8d181302356
    # changes requests.adapters.HTTPAdapter to no longer call get_connection() from
    # send(), but instead call _get_connection().
    def _get_connection(self, request, *args, **kwargs):  # type: ignore
        return self.get_connection(request.url, kwargs.get("proxies"))

    # Fix for requests 2.32.2+:
    # https://github.com/psf/requests/commit/c98e4d133ef29c46a9b68cd783087218a8075e05
    def get_connection_with_tls_context(self, request, verify, proxies=None, cert=None):  # type: ignore
        return self.get_connection(request.url, proxies)
